"""This module manages long running commands

To start a python http server
```
cmd = j.tools.startupcmd.get("cmd_name")

cmd.start_cmd = "python3 -m http.server"

cmd.start()

```

Check if it is running

```
cmd.is_running()
```

Two types of executor:
- tmux(default)
- foreground

You can attach to a running process by specufying correct `name`, `ports`, `process_strings`, `process_strings_regex`.
If it matches any of the above you would be able to perform available on that process.


- Special cases

you can add cmd.ports, cmd.process_strings_regex or cmd.process_strings_regex to reach the process pid
"""
from enum import Enum
from jumpscale.loader import j
from jumpscale.core.base import Base, fields

import time
from psutil import NoSuchProcess


class Executor(Enum):
    TMUX = "tmux"
    FOREGROUND = "foreground"


class StartupCmd(Base):
    start_cmd = fields.String()
    ports = fields.List(fields.Integer())
    executor = fields.Enum(Executor)
    check_cmd = fields.String()
    path = fields.String(default=j.core.dirs.TMPDIR)
    stop_cmd = fields.String()
    env = fields.Typed(dict, default={})
    timeout = fields.Integer(default=60)
    process_strings = fields.List(fields.String())
    process_strings_regex = fields.List(fields.String())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._process = None
        self._pid = None
        self._cmd_path = None
        self.__tmux_window = None

    def reset(self):
        self._process = None
        self._pid = None

    @property
    def pid(self):
        if not self._pid:
            pids = j.sals.process.get_pids(f"startupcmd_{self.instance_name}")
            if pids:
                self._pid = pids[0]
        return self._pid

    @property
    def cmd_path(self):
        if not self._cmd_path:
            self._cmd_path = j.sals.fs.join_paths(j.core.dirs.VARDIR, "cmds", f"{self.instance_name}.sh")
            j.sals.fs.mkdirs(j.sals.fs.dirname(self._cmd_path))
        return self._cmd_path

    @pid.setter
    def pid(self, pid):
        self._pid = pid

    @property
    def process(self):
        if not self._process:
            if self.pid:
                self._process = j.sals.process.get_process_object(self.pid, die=False)
                if not self._process:
                    self.pid = None
            else:
                processes = self._get_processes_by_port_or_filter()
                if len(processes) == 1:
                    self._process = processes[0]
        return self._process

    @property
    def _tmux_window(self):
        if self.executor.value == Executor.TMUX.value:
            if self.__tmux_window is None:
                self.__tmux_window = j.core.executors.tmux.get_js_window(self.instance_name)
        return self.__tmux_window

    def _get_processes_by_port_or_filter(self):
        """Uses object properties to find the corresponding process(es)

        Returns:
            list: All processes that matched
        """

        pids_done = []
        result = []

        def _add_to_result(process):
            if process and process.pid not in pids_done:
                result.append(process)
                pids_done.append(process.pid)

        for port in self.ports:
            try:
                process = j.sals.process.get_process_by_port(port)
            except Exception:
                continue

            _add_to_result(process)

        for process_string in self.process_strings:
            for pid in j.sals.process.get_filtered_pids(process_string):
                process = j.sals.process.get_process_object(pid, die=False)
                _add_to_result(process)

        for pid in j.sals.process.get_pids_filtered_by_regex(self.process_strings_regex):
            process = j.sals.process.get_process_object(pid, die=False)
            _add_to_result(process)

        #  We return all processes which match
        return result

    def _kill_processes_by_port_or_filter(self):
        """Kills processes that matches object properties
        """
        processes = self._get_processes_by_port_or_filter()
        self._kill_processes(processes)

    def _kill_processes(self, processes):
        """Kill processes

        Args:
            processes (list): List of processes
        """
        for process in processes:
            try:
                process.kill()
            except NoSuchProcess:
                pass  # already killed

    def _soft_kill(self):
        """Kills the poocess using `stop_cmd`

        Returns:
            bool: True if was killed
        """
        if self.stop_cmd:
            cmd = j.tools.jinja2.render_template(template_text=self.stop_cmd, args=self._get_data())
            exit_code, _, _ = j.sals.process.execute(cmd, die=False)
            self.reset()
            return exit_code == 0
        elif self.process:
            try:
                self.process.terminate()
                return True
            except Exception:
                pass
        return False

    def _hard_kill(self):
        """Force Kills the process
        """
        if self.process:
            self._kill_processes([self.process])
            self.reset()

        self._kill_processes_by_port_or_filter()

        if self.executor.value == Executor.TMUX.value:
            self._tmux_window.kill_window()
            self.__tmux_window = None

    def stop(self, force=True, wait_for_stop=True, die=True, timeout=None):
        """Stops the running command

        Args:
            force (bool, optional): If True will force kill the process. Defaults to True.
            wait_for_stop (bool, optional): If True will wait until process is stopped. Defaults to True.
            die (bool, optional): If True will raise if timeout is exceeded for stop. Defaults to True.
            timeout (int, optional): Timeout for stop wait.If not set will use `timeout` property. Defaults to None.
        """

        timeout = timeout or self.timeout

        if self.is_running():
            if self._soft_kill():
                self.wait_for_stop(die=False, timeout=timeout)

        if force:
            self._hard_kill()

        if wait_for_stop:
            self.wait_for_stop(die=die, timeout=timeout)
        j.sals.process.execute(f"rm {self.cmd_path}", die=False)

    def is_running(self):
        """Checks if startup cmd is running. Will use `check_cmd` property if defined or check based on objet properties

        Returns:
            bool: True if it is running
        """
        if self.check_cmd:
            exit_code, _, _ = j.sals.process.execute(self.check_cmd, die=False)
            return exit_code == 0

        self.reset()
        if self.process:
            return self.process.is_running()
        return self._get_processes_by_port_or_filter() != []

    def _wait(self, for_running, die, timeout):
        """Wait for either start or stop to finishes

        Args:
            for_running (bool): Whether to check if it is running or stopped.
            die (bool, optional): If True will raise if timeout is exceeded for stop. Defaults to True.
            timeout (int, optional): Timeout for wait operation. Defaults to None.

        Raises:
            j.exceptions.Timeout: If timeout is exceeded.
        """
        end = j.data.time.now().timestamp + timeout

        while j.data.time.now().timestamp < end:
            if self.is_running() == for_running:
                break
            time.sleep(0.05)
        else:
            if die:
                raise j.exceptions.Timeout(f"Wait operation exceeded timeout: {timeout}")

    def wait_for_stop(self, die=True, timeout=10):
        """Wait for stop to finishes

        Args:
            die (bool, optional): If True will raise if timeout is exceeded for stop. Defaults to True.
            timeout (int, optional): Timeout for wait operation. Defaults to None.

        Raises:
            j.exceptions.Timeout: If timeout is exceeded.
        """
        self._wait(False, die, timeout)

    def wait_for_running(self, die=True, timeout=10):
        """Wait for start to finishes

        Args:
            die (bool, optional): If True will raise if timeout is exceeded for stop. Defaults to True.
            timeout (int, optional): Timeout for wait operation. Defaults to None.

        Raises:
            j.exceptions.Timeout: If timeout is exceeded.
        """
        self._wait(True, die, timeout)

    def start(self):
        """Starts the process
        """
        if self.is_running():
            return

        if not self.start_cmd:
            raise j.exceptions.Value("please make sure start_cmd has been set")

        if "\n" in self.start_cmd.strip():
            command = self.start_cmd
        else:
            template_script = """
            set +ex
            {% for key,val in env.items() %}
            export {{key}}='{{val}}'
            {% endfor %}

            mkdir -p {{path}}
            cd {{path}}
            bash -c \"exec -a startupcmd_{{name}} {{start_cmd}}\"

            """

            script = j.tools.jinja2.render_template(
                template_text=template_script,
                env=self.env,
                path=self.path,
                start_cmd=self.start_cmd,
                name=self.instance_name,
            )
            j.sals.fs.write_file(self.cmd_path, script)
            j.sals.fs.chmod(self.cmd_path, 0o770)
            command = f"sh {self.cmd_path}"

        if self.executor.value == Executor.FOREGROUND.value:
            j.sals.process.execute(command)
        elif self.executor.value == Executor.TMUX.value:
            self._tmux_window.attached_pane.send_keys(command)

        self.wait_for_running(die=True, timeout=self.timeout)



from jumpscale.data.serializers import json


class Serializer:
    def serialize(self, obj):
        return obj

    def deserialize(self, data):
        return data


class JsonSerializer(Serializer):
    def serialize(self, obj):
        return json.dumps(obj)

    def deserialize(self, data):
        return json.loads(data)



from configparser import ConfigParser


class IniFile:
    """
    the IniFile object parses the content of the file provided by the path argument
    Args:
        path (str) : the path to file.ini
    """

    def __init__(self, path):
        self.path = path
        self.parser = ConfigParser()
        self.parser.read(self.path)

    def write(self):
        """
        apply all the changes to the file
        """
        with open(self.path, "w") as file:
            self.parser.write(file)

    def get_sections(self):
        """
        search the file for sections names
        Returns:
            a list of the sections
        """
        return self.parser.sections()

    def get_properties(self, section_name):
        """
        get the properties name under the provided section
        Args:
            section_name (str) : the section name which contains the properties
        Returns:
            list of all the properties under the section
        """
        return self.parser.options(section_name)

    def check_section(self, section_name):
        """
        check the existence of the section
        Args:
            section_name (str) : the section wanted to check
        Returns:
            boolen expresion
        """
        return section_name in self.parser.sections()

    def check_property(self, section_name, property_name):
        """
        check the existence of the property in section
        Args:
            section_name (str) : the name of the section where the property is
            property_name (str) : the property wanted to check
        Returns:
            boolen expresion
        """
        return property_name in self.parser.options(section_name)

    def get_value(self, section_name, property_name):
        """
        gat the value of property as string
        Args:
            section_name (str) : the name of the section where the property is
            property_name (str) : the property wanted to get its value
        Returns:
            string of the value
        """
        return self.parser.get(section_name, property_name)

    def get_int(self, section_name, property_name):
        """
        gat the value of property as int
        Args:
            section_name (str) : the name of the section where the property is
            property_name (str) : the property wanted to get its value
        Returns:
            int of the value
        """
        return self.parser.getint(section_name, property_name)

    def get_float(self, section_name, property_name):
        """
        gat the value of property as float
        Args:
            section_name (str) : the name of the section where the property is
            property_name (str) : the property wanted to get its value
        Returns:
            float of the value
        """
        return self.parser.getfloat(section_name, property_name)

    def get_boolen(self, section_name, property_name):
        """
        gat the value of property as boolen
        Args:
            section_name (str) : the name of the section where the property is
            property_name (str) : the property wanted to get its value
        Returns:
            boolen of the value
        """
        return self.parser.getboolean(section_name, property_name)

    def add_section(self, section_name):
        """
        add an empty section to the file
        Args:
            section_name (str) : the section name
        """
        self.parser[section_name] = {}

    def add_property(self, section_name, property_key, property_value):
        """
        add new property to the file
        Args:
            section_name (str) : the section name
            property_key (str) : the name of the property
            property_value (str) : the value of the property
        """
        self.parser[section_name] = {property_key: property_value}

    def remove_section(self, section_name):
        """
        delete a section and its properties
        Args:
            section_name (str) : the section name
        """
        self.parser.remove_section(section_name)

    def remove_property(self, section_name, property_name):
        """
        delete a property from section
        Args:
            section_name (str) : the section name
            property_key (str) : the name of the property
        """
        self.parser.remove_option(section_name, property_name)




"""
Docs for nettools

"""
# TODO: test *2

import time
import socket
import ipaddress
import re
from typing import Optional
from jumpscale.core.exceptions import Value, Runtime, Input
from jumpscale.data.time import now
import jumpscale.tools.http
import jumpscale.data.platform
import jumpscale.sals.fs
import jumpscale.core.executors
from jumpscale.data.types import IPAddress


def tcp_connection_test(ipaddr: str, port: int, timeout: Optional[int]):
    """tests tcp connection on specified port

    Args:
        ipaddr (str): ip address
        port (int): port number
        timeout (int, optional): time before the connection test fails. Defaults to None.

    Returns:
        bool: True if the test succeeds, False otherwise
    """
    conn = None
    try:
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if timeout:
            conn.settimeout(timeout)
        try:
            conn.connect((ipaddr, port))
        except BaseException:
            return False
    finally:
        if conn:
            conn.close()
    return True


def udp_connection_test(ipaddr: str, port: int, timeout=1, message=b"PING"):
    """tests udp connection on specified port by sending specified message

    Args:
        ipaddr (str): ip address
        port (int): port number
        timeout (int, optional): time before the connection test fails. Defaults to None.
        message (str, optional): message to send. Defaults to b"PING"

    Returns:
        bool: True if the test succeeds, False otherwise
    """
    conn = socket.socket(type=socket.SOCK_DGRAM)
    if timeout:
        conn.settimeout(timeout)
    try:
        conn.connect((ipaddr, port))
    except BaseException:
        conn.close()
        return False

    conn.send(message)

    try:
        conn.recvfrom(8192)
    except Exception:
        return False
    return True


def wait_connection_test(ipaddr: str, port: int, timeout=5):
    """Will wait until port listens on the specified address

    Args:
        ipaddr (str): ip address
        port (int): port number
        timeout_total (int, optional): how long to wait for the connection. Defaults to 5.

    Returns:
        bool: True if the test succeeds, False otherwise
    """
    port = int(port)
    end = now().timestamp + timeout
    while True:
        if now().timestamp > end:
            return False
        if tcp_connection_test(ipaddr, port, timeout=2):
            return True


def check_url_reachabel(url: str, timeout=5):
    """Check that given url is reachable

    Args:
        url (str): url to test
        timeout (int, optional): timeout of test. Defaults to 5.

    Raises:
        Input: raises if not correct url

    Returns:
        bool: True if the test succeeds, False otherwise
    """
    try:
        code = jumpscale.tools.http.get(url, timeout=timeout).status_code
        return code == 200
    except jumpscale.tools.http.exceptions.MissingSchema:
        raise Input("Please specify correct url with correct scheme")
    except jumpscale.tools.http.exceptions.ConnectionError:
        return False


def get_nic_names():
    """Get Nics on this machine

    Returns:
        list: list of all availabe nics
    """
    return [nic["name"] for nic in get_network_info()]


def get_nic_type(interface):
    """Get Nic Type on a certain interface

    Args:
        interface (str): interface name

    Raises:
        Runtime: if ethtool not installed on the system
        Value: if interface given is invalid

    Returns:
        str: type of the interface
    """
    output = ""
    if jumpscale.data.platform.is_linux():
        if jumpscale.sals.fs.exists(f"/sys/class/net/{interface}"):
            output = jumpscale.sals.fs.read_file(f"/sys/class/net/{interface}/type")
        if output.strip() == "32":
            return "INFINIBAND"
        else:
            if jumpscale.sals.fs.exists("/proc/net/vlan/%s" % (interface)):
                return "VLAN"
            exitcode, _, _ = jumpscale.core.executors.run_local("which ethtool", hide=True, warn=True)
            if exitcode != 0:
                raise Runtime("Ethtool is not installed on this system!")
            exitcode, output, _ = jumpscale.core.executors.run_local(f"ethtool -i {interface}", hide=True, warn=True)
            if exitcode != 0:
                return "VIRTUAL"
            match = re.search(r"^driver:\s+(?P<driver>\w+)\s*$", output, re.MULTILINE)
            if match and match.group("driver") == "tun":
                return "VIRTUAL"
            if match and match.group("driver") == "bridge":
                return "VLAN"
            return "ETHERNET_GB"

    elif jumpscale.data.platform.is_osx():
        command = f"ifconfig {interface}"
        exitcode, output, err = jumpscale.core.executors.run_local(command, hide=True, warn=True)
        if exitcode != 0:
            # temporary plumb the interface to lookup its mac
            jumpscale.core.executors.run_local(f"{command} plumb", hide=True)
            exitcode, output, err = jumpscale.core.executors.run_local(command, hide=True)
            jumpscale.core.executors.run_local(f"{command} unplumb", hide=True)
        if output.find("ipib") >= 0:
            return "INFINIBAND"
        else:
            # work with interfaces which are subnetted on vlans eq e1000g5000:1
            interfacepieces = interface.split(":")
            interface = interfacepieces[0]
            match = re.search(r"^\w+?(?P<interfaceid>\d+)$", interface, re.MULTILINE)
            if not match:
                raise Value(f"Invalid interface {interface}")
            if len(match.group("interfaceid")) >= 4:
                return "VLAN"
            else:
                if len(interfacepieces) > 1:
                    return "VIRTUAL"
                else:
                    return "ETHERNET_GB"


def get_reachable_ip_address(ip: str, port: int):
    """Returns the first local ip address that can connect to the specified ip on the specified port

    Args:
        ip (str): ip address
        port (int): port number

    Raises:
        Runtime: if can't connect

    Returns:
        str: ip that can connect to the specified ip
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect((ip, port))
    except BaseException:
        raise Runtime("Cannot connect to %s:%s, check network configuration" % (ip, port))
    return s.getsockname()[0]


def get_default_ip_config():
    """get default nic and address

    Returns:
        tuple: default nic and address
    """
    ipaddr = get_reachable_ip_address("8.8.8.8", 22)
    for item in get_network_info():
        for ipaddr2 in item["ip"]:
            # print "%s %s"%(ipaddr2,ipaddr)
            if str(ipaddr) == str(ipaddr2):
                return item["name"], ipaddr


def get_network_info(device=None):
    """
    Get network info

    [{'cidr': 8, 'ip': ['127.0.0.1'], 'mac': '00:00:00:00:00:00', 'name': 'lo'},
        {'cidr': 24,
        'ip': ['192.168.0.105'],
        'ip6': ['...','...],
        'mac': '80:ee:73:a9:19:05',
        'name': 'enp2s0'},
        {'cidr': 0, 'ip': [], 'mac': '80:ee:73:a9:19:06', 'name': 'enp3s0'},
        {'cidr': 16,
        'ip': ['172.17.0.1'],
        'mac': '02:42:97:63:e6:ba',
        'name': 'docker0'}]

    :param device: device name, defaults to None
    :type device: str, optional
    :raises RuntimeError: if the platform isn't implemented
    :return: network info
    :rtype: list or dict if device is specified
    """
    IPBLOCKS = re.compile(r"(^|\n)(?P<block>\d+:.*?)(?=(\n\d+)|$)", re.S)
    IPMAC = re.compile(r"^\s+link/\w+\s+(?P<mac>(\w+:){5}\w{2})", re.M)
    IPIP = re.compile(r"\s+?inet\s(?P<ip>(\d+\.){3}\d+)/(?P<cidr>\d+)", re.M)
    IPNAME = re.compile(r"^\d+: (?P<name>.*?)(?=:)", re.M)

    def block_parse(block):
        result = {"ip": [], "ip6": [], "cidr": [], "mac": "", "name": ""}
        for rec in (IPMAC, IPNAME):
            match = rec.search(block)
            if match:
                result.update(match.groupdict())
        for mrec in (IPIP,):
            for m in mrec.finditer(block):
                for key, value in list(m.groupdict().items()):
                    result[key].append(value)
        _, IPV6, _ = jumpscale.core.executors.run_local(
            "ifconfig %s |  awk '/inet6/{print $2}'" % result["name"], hide=True
        )
        for ipv6 in IPV6.split("\n"):
            result["ip6"].append(ipv6)
        if isinstance(result["cidr"], list):
            if len(result["cidr"]) == 0:
                result["cidr"] = 0
            else:
                result["cidr"] = int(result["cidr"][0])
        return result

    def networkinfo_get():
        _, output, _ = jumpscale.core.executors.run_local("ip a", hide=True)
        for m in IPBLOCKS.finditer(output):
            block = m.group("block")
            yield block_parse(block)

    res = []
    for nic in networkinfo_get():
        if nic["name"] == device:
            return nic
        res.append(nic)

    if device is not None:
        raise Runtime("could not find device")
    return res


def get_mac_address(interface: str):
    """Return the MAC address of this interface

    Args:
        interface (str): interface name

    Returns:
        str: mac of the interface
    """
    return get_network_info(interface)["mac"]


def get_host_name():
    """Get hostname of the machine

    Returns:
        str: host name
    """
    return socket.gethostname()


def is_nic_connected(interface: str):
    """check if interface is connected

    Args:
        interface (str): interface name

    Returns:
        bool: whether it is connected or not
    """
    if jumpscale.data.platform.is_linux():
        carrierfile = f"/sys/class/net/{interface}/carrier"
        if not jumpscale.sals.fs.exists(carrierfile):
            return False
        try:
            return int(jumpscale.sals.fs.read_file(carrierfile)) != 0
        except IOError:
            return False

    elif jumpscale.data.platform.is_osx():
        command = f"dladm show-dev -p -o STATE {interface}"
        expectResults = ["up", "unknown"]

        exitcode, output, _ = jumpscale.core.executors.run_local(command, warn=True, hide=True)
        if exitcode != 0:
            return False
        output = output.strip()
        return output in expectResults


def get_host_by_name(dnsHostname: str):
    """get host address by its name

    Args:
        dnsHostname (str): host name

    Returns:
        str: host address
    """
    return socket.gethostbyname(dnsHostname)


def ping_machine(ip, pingtimeout=60, allowhostname=True):
    """Ping a machine to check if it's up/running and accessible
    @param ip: Machine Ip Address
    @param pingtimeout: time in sec after which ip will be declared as unreachable
    @param allowhostname: allow pinging on hostname (default is false)
    @rtype: True if machine is pingable, False otherwise
    """
    if not allowhostname:
        if not IPAddress().check(ip):
            raise Value("Invalid ip address, set allowedhostname to use hostnames")

    start = time.time()
    pingsucceeded = False
    while time.time() - start < pingtimeout:
        if jumpscale.data.platform.is_linux():
            # ping -c 1 -W 1 IP
            exitcode, _, _ = jumpscale.core.executors.run_local(f"ping -c 1 -W 1 -w 1 {ip}", warn=True, hide=True)
        elif jumpscale.data.platform.is_osx():
            exitcode, _, _ = jumpscale.core.executors.run_local(f"ping -c 1 {ip}", warn=True, hide=True)
        if exitcode == 0:
            pingsucceeded = True
            return True
        time.sleep(1)
    if not pingsucceeded:
        return False


def download(url, localpath, username=None, passwd=None, overwrite=True):
    """Download a url to a file or a directory, supported protocols: http, https, ftp, file
    @param url: URL to download from
    @type url: string
    @param localpath: filename or directory to download the url to pass - to return data
    @type localpath: string
    @param username: username for the url if it requires authentication
    @type username: string
    @param passwd: password for the url if it requires authentication
    @type passwd: string
    """
    if not url:
        raise Value("URL can not be None or empty string")
    if not localpath:
        raise Value("Local path to download the url to can not be None or empty string")
    filename = ""
    if localpath == "-":
        filename = "-"
    if jumpscale.sals.fs.exists(localpath) and jumpscale.sals.fs.is_dir(localpath):
        filename = jumpscale.sals.fs.join_paths(localpath, jumpscale.sals.fs.basename(url))
    else:
        if jumpscale.sals.fs.is_dir(jumpscale.sals.fs.dirname(localpath)):
            filename = localpath
        else:
            raise Value("Local path is an invalid path")

    if not jumpscale.sals.fs.exists(filename):
        overwrite = True

    if overwrite:
        if username and passwd and jumpscale.tools.http.urllib3.util.parse_url(url).scheme == "ftp":
            url = url.split("://")[0] + "://%s:%s@" % (username, passwd) + url.split("://")[1]
        response = jumpscale.tools.http.get(
            url, stream=True, auth=jumpscale.tools.http.auth.HTTPBasicAuth(username, passwd)
        )
        response.raise_for_status()
        if filename != "-":
            with open(filename, "wb") as fw:
                for chunk in response.iter_content(chunk_size=2048):
                    if chunk:  # filter out keep-alive new chunks
                        fw.write(chunk)
            return
        else:
            return response.content


def _netobject_get(device: str):
    n = get_network_info(device)
    net = ipaddress.IPv4Network(n["ip"][0] + "/" + str(n["cidr"]), strict=False)
    return net


def netrange_get(device, skip_begin=10, skip_end=10):
    """
    Get ($fromip,$topip) from range attached to device, skip the mentioned ip addresses.

    :param device: device name
    :type device: str
    :param skip_begin: ips to skip from the begining of the range, defaults to 10
    :type skip_begin: int, optional
    :param skip_end: ips to skip from the end of the range, defaults to 10
    :type skip_end: int, optional

    :return: ip range for this device
    :rtype: tuple
    """
    n = _netobject_get(device)
    return (str(n[0] + skip_begin), str(n[-1] - skip_end))



from functools import wraps
from subprocess import list2cmdline


def format_cmd(cmd):
    ## code around it to run in tmux.
    pass


def cmd_from_args(func):
    """
    a decorator to allow passing cmd as a list, with auto-escaping using `subprocess.list2cmdline`

    cmd must be the first positional arguments

    Args:
        func (function): the function to decorate

    Returns:
        function: a new function
    """

    @wraps(func)
    def inner(cmd, *args, **kwargs):
        if isinstance(cmd, list):
            cmd = list2cmdline(cmd)

        return func(cmd, *args, **kwargs)

    return inner



import os
import sys
import json
import inspect
from jumpscale.loader import j
from jumpscale.servers.gedis.baseactor import BaseActor, actor_method


class CoreActor(BaseActor):
    def __init__(self):
        super().__init__()
        self._server = None
        self.path = __file__

    def set_server(self, server):
        self._server = server

    @actor_method
    def list_actors(self) -> list:
        """List available actors

        Returns:
            list -- list of available actors
        """
        return list(self._server._loaded_actors.keys())


class SystemActor(BaseActor):
    def __init__(self):
        super().__init__()
        self._server = None
        self.path = __file__

    def set_server(self, server):
        self._server = server

    @actor_method
    def register_actor(self, actor_name: str, actor_path: str) -> bool:
        """Register new actor

        Arguments:
            actor_name {str} -- actor name within gedis server.
            actor_path {str} -- actor path on gedis server machine.

        Returns:
            bool -- True if registered.
        """
        module = j.tools.codeloader.load_python_module(actor_path)
        actor = module.Actor()
        actor.path = actor_path
        result = actor.__validate_actor__()

        if not result["valid"]:
            raise j.exceptions.Validation(
                "Actor {} is not valid, check the following errors {}".format(actor_name, result["errors"])
            )

        self._server._register_actor(actor_name, actor)
        return True

    @actor_method
    def unregister_actor(self, actor_name: str) -> bool:
        """Register actor

        Arguments:
            actor_name {str} -- actor name

        Returns:
            bool -- True if actors is unregistered
        """
        self._server._unregister_actor(actor_name)
        return True



'''Highlighter module helps with formatting text to be highlighted in terminal and in web

example:

```
def test():
    C = """
    def _init(self,**kwargs):
        self.lexers = Lexers()
        self.formatters = Formatters()
        

    def print_python(self,text,formatter="terminal"):
        C=Tools.text_replace(text)
        print(pygments.highlight(C,self.lexers.get("python"), self.formatters.get(formatter)))


    """

    print_python(C)
    print_highlighted(C)

    print("####TOML EXAMPLE####")

    C = """

    title = "TOML Example"

    [owner]
    name = "Tom Preston-Werner"
    organization = "GitHub"
    bio = "GitHub Cofounder & CEO"
    dob = 1979-05-27T07:32:00Z # First class dates? Why not?

    [database]
    server = "192.168.1.1"
    ports = [ 8001, 8001, 8002 ]
    connection_max = 5000
    enabled = true
            
    """
    print_toml(C)
    print_highlighted(C)
```
'''
import pygments
from pygments.formatters import get_formatter_by_name
from pygments.lexers import guess_lexer, find_lexer_class_by_name


def print_highlighted(txt, lexer=None, formatter="terminal"):
    lexer = find_lexer_class_by_name(lexer) if lexer else guess_lexer(txt)
    print(pygments.highlight(txt, lexer, get_formatter_by_name(formatter)))


print_python = print_highlighted
print_toml = print_highlighted


def test():
    C = """
    def _init(self,**kwargs):
        self.lexers = Lexers()
        self.formatters = Formatters()
        

    def print_python(self,text,formatter="terminal"):
        C=Tools.text_replace(text)
        print(pygments.highlight(C,self.lexers.get("python"), self.formatters.get(formatter)))


    """

    print_python(C)
    print_highlighted(C)

    print("####TOML EXAMPLE####")

    C = """

    title = "TOML Example"

    [owner]
    name = "Tom Preston-Werner"
    organization = "GitHub"
    bio = "GitHub Cofounder & CEO"
    dob = 1979-05-27T07:32:00Z # First class dates? Why not?

    [database]
    server = "192.168.1.1"
    ports = [ 8001, 8001, 8002 ]
    connection_max = 5000
    enabled = true
            
    """
    print_toml(C)
    print_highlighted(C)



"""Gedis server package provides all code needed to have an RPC server using redis protocol for messaging

### Create Gedis instance
```
t = j.servers.gedis.get("test")
```
this will create Gedis instance with name test running on  default port 16000
```
t = j.servers.gedis.new("test",port=1500)
```
This will create Gedis instance with name test running on port 1500
### Start Gedis server
```
t.start()
```
### Stop Gedis server
```
t.stop()
```

~>  redis-cli -p 16000 greeter hi
actor greeter isn't loaded
 ~>  redis-cli -p 16000 system register_actor greeter /home/ahmed/wspace/threefoldtech/js-ng/jumpscale/servers/gedis/example_greeter.py
(integer) -1
 ~>  redis-cli -p 16000 greeter hi
 hello world
 ~>  redis-cli -p 16000 greeter add2 jo deboeck
 "jodeboeck"
 ~>  fuser -k 16000/tcp

16000/tcp:           29331
 ~>  redis-cli -p 16000 greeter hi
 actor greeter isn't loaded
 ~>  redis-cli -p 16000 system register_actor greeter /home/ahmed/wspace/threefoldtech/js-ng/jumpscale/servers/gedis/example_greeter.py
(integer) -1
 ~>  redis-cli -p 16000 greeter hi
 hello world
 ~>  redis-cli -p 16000 greeter ping

pong no?
 ~>  redis-cli -p 16000 greeter add2 reem khamis
"reemkhamis"
```
"""
from jumpscale.core.base import StoredFactory


def export_module_as():
    from .server import GedisServer

    return StoredFactory(GedisServer)



from jumpscale.servers.gedis.baseactor import BaseActor, actor_method
from typing import Sequence
from jumpscale.loader import j
import inspect, sys


class TestObject:
    def __init__(self):
        self.attr = None

    def to_dict(self):
        return self.__dict__

    def from_dict(self, ddict):
        self.__dict__ = ddict


class Example(BaseActor):
    @actor_method
    def add_two_ints(self, x: int, y: int) -> int:
        """Adds two ints

        Arguments:
            x {int} -- first int
            y {int} -- second int

        Returns:
            int -- the sum of the two ints
        """
        return x + y

    @actor_method
    def concate_two_strings(self, x: str, y: str) -> str:
        """Concate two strings

        Arguments:
            x {str} -- first string
            y {str} -- second string

        Returns:
            str -- the concate of the two strings
        """
        return x + y

    @actor_method
    def modify_object(self, myobj: list, new_value: int) -> list:
        """Modify atrribute attr of the given object

        Arguments:
            myobj {TestObject} -- the object to be modified

        Returns:
            TestObject -- modified object
        """
        for i in range(len(myobj)):
            myobj[i].attr = new_value * (i + 1)
        return myobj


Actor = Example



"""Git module helps with everything around git management like pulling, cloning .. etc

# TODO: examples
"""
import urllib
import re
from jumpscale.loader import j

SSH_URL_MATCH = "^(git@)(?P<netloc>.*?)(:|/)(?P<path>.*?)/?$"


def rewrite_git_https_url_to_ssh(url):
    """
    :param url: (str) https url to rewrite as ssh
    :return (tuple) repo name and rewritten ssh url
    """
    parsed_url = urllib.parse.urlparse(url)
    if parsed_url.scheme != "https":
        raise j.exceptions.Input("not a https url: {}".format(url))
    return "git@{}:{}.git".format(parsed_url.netloc, parsed_url.path.lstrip("/"))


def rewrite_git_ssh_url_to_https(url, login=None, passwd=None):
    """
    :param url: (str) ssh url to rewrite ass https
    :param login: (str) authentication login name
    :param passwd: (str) authentication login password
    :return (tuple) repo name and rewritten https url
    """
    match = re.match(SSH_URL_MATCH, url)
    url_data = match.groupdict()
    if not match:
        raise j.exceptions.Input("not a ssh url: {}".format(url))
    if (login or passwd) and not (login and passwd):
        raise j.exceptions.Input("Need to specify both login and passwd if either one is specified")
    elif login and passwd:
        rewrite_url = "https://{login}:{passwd}@{netloc}/{path}".format(**match.groupdict(), login=login, passwd=passwd)
    else:
        rewrite_url = "https://{netloc}/{path}".format(**url_data)
    return rewrite_url


def ensure_repo(url: str, dest="", branch_or_tag="", commit_id="", discard_changes=False, depth=0):
    """Makes sure that repo exists in specified dest with correct branch

    Args:
        url (str): url of the repo
        dest (str, optional): the repo path, if it doesn't exist will clone it. Will get path from url if not specified.
        branch_or_tag (str, optional): clone from a specific branch or tag. Defaults to "".
        commit_id (str, optional): commit to checkout to. Defaults to "".
        discard_changes (bool, optional): Will remove changes in repo if True during pull. Defaults to False.
        depth (int, optional): specify the depth of the clone. Defaults to 0.
    """
    dest = dest or get_path_from_url(url)
    if j.sals.fs.exists(dest):
        revision = branch_or_tag or commit_id
        pull_repo(dest, discard_changes, revision)
    else:
        parent_dir = str(j.sals.fs.parent(dest))
        clone_repo(url, parent_dir, branch_or_tag, depth)
    return dest


def clone_repo(url: str, dest: str, branch_or_tag="", depth=0, commit_id=""):
    """Clones repo with sepcified url at specified dest

    Args:
        url (str): url of the repo
        dest (str): destination to clone in
        branch_or_tag (str, optional): clone from a specific branch or tag. Defaults to "".
        depth (int, optional): specify the depth of the clone. Defaults to 0.
        commit_id (int, optional): commit to checkout to. Defauls to "".
    """
    j.sals.fs.mkdirs(dest)
    prefix = f"cd {dest} && "
    cmd = prefix + f"git clone --single-branch {url}"
    if branch_or_tag:
        cmd += f" -b {branch_or_tag}"
    if depth != 0:
        cmd += f" --depth={depth}"
    rc, _, err = j.core.executors.run_local(cmd, warn=True)
    if rc > 0:
        raise j.exceptions.Runtime(f"Error in execute {cmd}\n{err}")
    repo_name = j.sals.fs.basename(url).split(".git")[0]
    if commit_id:
        prefix = f"cd {dest}/{repo_name} && "
        checkout_cmd = prefix + f"git checkout {commit_id}"
        rc, _, err = j.core.executors.run_local(checkout_cmd, warn=True)
        if rc > 0:
            raise j.exceptions.Runtime(f"Error in execute {checkout_cmd}\n{err}")
    return repo_name


def pull_repo(path: str, discard_changes=False, revision=""):
    """Pull repo at the given path

    Args:
        path (str): path of the git repo to pull.
        discard_changes (bool, optional): Will remove changes in repo if True. Defaults to False.
        revision (str, optional): change to specified branch, tag, commit. Defaults to "".
    """
    prefix = f"cd {path} && "
    if discard_changes:
        j.core.executors.run_local(prefix + "git checkout .")
    j.core.executors.run_local(prefix + "git pull")
    if revision:
        j.core.executors.run_local(prefix + f"git checkout {revision}")


def giturl_parse(url):
    """
    :return
    - repository_url the full url to the repo but rewritten
    - relpath: path inside the git repo
    - branch: branch specified in the url

    example Input
    - https://github.com/threefoldtech/jumpscale_/NOS/blob/master/specs/NOS_1.0.0.md
    - https://github.com/threefoldtech/jumpscale_/jumpscaleX_core/blob/8.1.2/lib/Jumpscale/tools/docsite/macros/dot.py
    - https://github.com/threefoldtech/jumpscale_/jumpscaleX_core/tree/8.2.0/lib/Jumpscale/tools/docsite/macros
    - https://github.com/threefoldtech/jumpscale_/jumpscaleX_core/tree/master/lib/Jumpscale/tools/docsite/macros

    """
    url_end = None
    branch = path = ""
    if "tree" in url:
        url, url_end = url.split("tree")
    elif "blob" in url:
        url, url_end = url.split("blob")
    if url_end is not None:
        url_end = url_end.strip("/")
        if url_end.find("/") == -1:
            path = ""
            branch = url_end
            if branch.endswith(".git"):
                branch = branch[:-4]
        else:
            branch, path = url_end.split("/", 1)
            if path.endswith(".git"):
                path = path[:-4]

    return url, path, branch


def get_path_from_url(url: str):
    """returns repo path in the filesystem based on the urk

    Args:
        url (str): repo url
    """

    def _get_path(netloc, path):
        return f"{j.core.dirs.CODEDIR}/{netloc.split('.')[0]}/{path.split('.git')[0]}"

    match = re.match(SSH_URL_MATCH, url)
    if match:
        url_data = match.groupdict()
        path = _get_path(**url_data)
    else:
        url = giturl_parse(url)[0]
        parsed_url = urllib.parse.urlparse(url)
        path = _get_path(parsed_url.netloc, parsed_url.path)
    return path


def find(account="", name=""):
    """find all repo paths in j.core.dirs.CODEDIR

    Args:
        account (str, optional): returns repos under this account only. Defaults to "".
        name (str, optional): retuns repos with this name only. Defaults to "".

    Returns:
        list: list of repos path
    """
    if name:
        name += "/.git"

    def _filter_paths(path):
        if f"{account}/{name}" in str(path):
            return True
        else:
            return False

    return [
        j.sals.fs.parent(path) for path in j.sals.fs.walk(j.core.dirs.CODEDIR, pat="*.git", filter_fun=_filter_paths)
    ]


def find_git_path(path, die=True):
    """
    given a path, check if this path or any of its parents is a git repo, return the first git repo
    :param path: (String) path from where to start search
    :returns (String) the first path which is a git repo
    :raises Exception when no git path can be found
    """
    while path != "":
        if j.sal.fs.exists(path=j.sal.fs.joinPaths(path, ".git")):
            return path
        path = j.sal.fs.getParent(path)
    if die:
        raise j.exceptions.Input("Cannot find git path in:%s" % path)



"""
# To get a new ZDB instance

zcl = j.clients.zdb.get("instance", admin=True, mode="user")

# This will gets an instance with admin capabilities in user mode

"""
from jumpscale.clients.base import Client
from jumpscale.core.base import fields
from enum import Enum


import redis
import struct


class Mode(Enum):
    SEQ = "seq"
    USER = "user"


class ZDBClient(Client):
    addr = fields.String(default="localhost")
    port = fields.Integer(default=9900)
    secret_ = fields.String(default="1234567")
    nsname = fields.String(default="test")
    admin = fields.Boolean(default=False)
    mode = fields.Enum(Mode)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # if not self.secret_:
        #     self.secret_ = j.core.myenv.adminsecret

        assert len(self.secret_) > 5

        if self.admin:
            self.nsname = "default"
        self.type = "ZDB"
        self._redis = None
        self.nsname = self.nsname.lower().strip()

        # if j.data.bcdb._master:
        #     self._model.trigger_add(self._update_trigger)

    def _update_trigger(self, obj, action, **kwargs):
        if action in ["save", "change"]:
            self._redis = None

    @property
    def redis(self):
        if not self._redis:
            pool = redis.ConnectionPool(
                host=self.addr,
                port=self.port,
                password=self.secret_,
                connection_class=ZDBConnection,
                namespace=self.nsname,
                namespace_password=self.secret_,
                admin=self.admin,
            )
            self._redis = _patch_redis_client(redis.Redis(connection_pool=pool))
        return self._redis

    def _key_encode(self, key):
        if self.mode.value == Mode.SEQ.value:
            if key is None:
                key = ""
            else:
                key = struct.pack("<I", key)
        return key

    def _key_decode(self, key):
        if self.mode.value == Mode.SEQ.value:
            key = struct.unpack("<I", key)[0]
        return key

    def set(self, data, key=None):
        key = key or ""
        key = self._key_encode(key)
        res = self.redis.execute_command("SET", key, data)
        if not res:
            return res
        return self._key_decode(res)

    def get(self, key):
        key = self._key_encode(key)
        return self.redis.execute_command("GET", key)

    def exists(self, key):
        key = self._key_encode(key)
        return self.redis.execute_command("EXISTS", key) == 1

    def delete(self, key):
        if not key:
            raise j.exceptions.Value("key must be provided")
        key = self._key_encode(key)
        self.redis.execute_command("DEL", key)

    def flush(self):
        """
        will remove all data from the database DANGEROUS !!!!
        This is only allowed on private and password protected namespace
        You need to select the namespace before running the command.
        :return:
        """
        if not self.nsname in ["default", "system"]:
            self.redis.execute_command("FLUSH")

    def stop(self):
        pass

    @property
    def nsinfo(self):
        cmd = self.redis.execute_command("NSINFO", self.nsname)
        return _parse_nsinfo(cmd.decode())

    def list(self, key_start=None, reverse=False):
        """
        list all the keys in the namespace

        :param key_start: if specified start to walk from that key instead of the first one, defaults to None
        :param key_start: str, optional
        :param reverse: decide how to walk the namespace
                        if False, walk from older to newer keys
                        if True walk from newer to older keys
                        defaults to False
        :param reverse: bool, optional
        :return: list of keys
        :rtype: [str]
        """
        result = []
        for key, data in self.iterate(key_start=key_start, reverse=reverse, keyonly=True):
            result.append(key)
        return result

    def iterate(self, key_start=None, reverse=False, keyonly=False):
        """
        walk over all the namespace and yield (key,data) for each entries in a namespace

        :param key_start: if specified start to walk from that key instead of the first one, defaults to None
        :param key_start: str, optional
        :param reverse: decide how to walk the namespace
                if False, walk from older to newer keys
                if True walk from newer to older keys
                defaults to False
        :param reverse: bool, optional
        :param keyonly: [description], defaults to False
        :param keyonly: bool, optional
        :raises e: [description]
        """

        next = None
        data = None

        if key_start is not None:
            next = self.redis.execute_command("KEYCUR", self._key_encode(key_start))
            if not keyonly:
                data = self.get(key_start)
            yield (key_start, data)

        CMD = "SCANX" if not reverse else "RSCAN"

        while True:
            try:
                if not next:
                    response = self.redis.execute_command(CMD)
                else:
                    response = self.redis.execute_command(CMD, next)
                # format of the response
                # see https://github.com/threefoldtech/0-db/tree/development#scan
            except redis.ResponseError as e:
                if e.args[0] == "No more data":
                    return
                raise e

            (next, results) = response
            for item in results:
                keyb, size, epoch = item
                key_new = self._key_decode(keyb)
                data = None
                if not keyonly:
                    data = self.redis.execute_command("GET", keyb)
                yield (key_new, data)

    @property
    def count(self):
        """
        :return: return the number of entries in the namespace
        :rtype: int
        """
        return self.nsinfo["entries"]

    def ping(self):
        """
        go to default namespace & ping
        :return:
        """
        return self.redis.ping()

    @property
    def next_id(self):
        """
        :return: return the next id
        :rtype: int
        """
        id_bytes = struct.pack("<I", int(self.nsinfo["next_internal_id"], 16))
        return int.from_bytes(id_bytes, byteorder="big", signed=True)


# Helper functions
def _patch_redis_client(redis):
    # don't auto parse response for set, it's not 100% redis compatible
    # 0-db does return a key after in set
    for cmd in ["SET", "DEL"]:
        if cmd in redis.response_callbacks:
            del redis.response_callbacks[cmd]
    return redis


def _parse_nsinfo(raw):
    def empty(line):
        line = line.strip()
        if len(line) <= 0 or line[0] == "#" or ":" not in line:
            return False
        return True

    info = {}
    for line in filter(empty, raw.splitlines()):
        key, val = line.split(":")
        try:
            val = int(val)
            info[key] = val
            continue
        except ValueError:
            pass
        try:
            val = float(val)
            info[key] = val
            continue
        except ValueError:
            pass

        info[key] = str(val).strip()
    return info


class ZDBConnection(redis.Connection):
    """
    ZDBConnection implement the custom selection of namespace
    on 0-DB
    """

    def __init__(self, namespace=None, namespace_password=None, admin=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.namespace = namespace or "default"
        self.namespace_password = namespace_password
        self.admin = admin

    def on_connect(self):
        """
        when a new connection is created, switch to the proper namespace
        """
        self._parser.on_connect(self)

        # if a password is specified, authenticate
        if self.admin and self.password:
            # avoid checking health here -- PING will fail if we try
            # to check the health prior to the AUTH
            self.send_command("AUTH", self.password, check_health=False)
            if redis.connection.nativestr(self.read_response()) != "OK":
                raise redis.connection.AuthenticationError("Invalid Password")

        # if a namespace is specified and it's not default, switch to it
        if self.namespace and not self.admin:
            args = [self.namespace]
            if self.namespace_password:
                args.append(self.namespace_password)

            self.send_command("SELECT", *args)
            if redis.connection.nativestr(self.read_response()) != "OK":
                raise redis.connection.ConnectionError(f"Failed to select namespace {self.namespace}")


class ZDBAdminClient(ZDBClient):
    def auth(self):
        assert self.admin
        if self.secret_:
            # authentication should only happen in zdbadmin client
            self.redis.execute_command("AUTH", self.secret_)

    def namespace_exists(self, name):
        assert self.admin
        self.auth()
        try:
            self.redis.execute_command("NSINFO", name)
            return True
        except Exception as e:
            if not "Namespace not found" in str(e):
                raise j.exceptions.Base("could not check namespace:%s, error:%s" % (name, e))
            return False

    def namespaces_list(self):
        assert self.admin
        self.auth()
        res = self.redis.execute_command("NSLIST")
        return [i.decode() for i in res]

    def namespace_new(self, name, secret=None, maxsize=0, die=False):
        """
        check namespace exists & will return zdb client to that namespace

        :param name:
        :param secret:
        :param maxsize:
        :param die:
        :return:
        """
        assert self.admin
        self.auth()
        if not self.namespace_exists(name):
            self.redis.execute_command("NSNEW", name)
        else:
            if die:
                raise j.exceptions.Base("namespace already exists:%s" % name)

        if secret:
            self.redis.execute_command("NSSET", name, "password", secret)
            self.redis.execute_command("NSSET", name, "public", "no")

        if maxsize != 0:
            self.redis.execute_command("NSSET", name, "maxsize", maxsize)

        ns = j.clients.zdb.client_get(
            name="temp_%s" % name, addr=self.addr, port=self.port, mode=self.mode, secret=secret, namespace=name
        )

        assert ns.ping()
        if secret:
            assert ns.nsinfo["public"] == "no"
        else:
            assert ns.nsinfo["public"] == "yes"

        ns._data.delete()
        return ns

    def namespace_get(self, name, secret=""):
        assert self.admin
        self.auth()
        return self.namespace_new(name, secret)

    def namespace_delete(self, name):
        assert self.admin
        self.auth()
        if self.namespace_exists(name):
            self.redis.execute_command("NSDEL", name)

    def reset(self, ignore=[]):
        """
        dangerous, will remove all namespaces & all data
        :param: list of namespace names not to reset
        :return:
        """
        assert self.admin
        self.auth()
        for name in self.namespaces_list():
            if name not in ["default"] and name not in ignore:
                self.namespace_delete(name)



"""
Meta and Base classes for any class with fields.

Contains mainly:

- `BaseMeta`: A meta class to get a new class with field property descriptors ready
- `Base`: The base class which can be used to get/set current field values


To explain what it does, we will illustrate the following examples:

If we have a class called `Person`, with the following definition:


```python
class Person:
    name = fields.String(default="ahmed")
```

Accessing name from class or instance level will yield the same value, an instance of `String` field:

```python
Person.name  #=> <jumpscale.core.base.fields.String object at 0x7efd89980c18>

p = Person()
p.name  #=> <jumpscale.core.base.fields.String object at 0x7efd89980c18>
```

The solution to this problem is using data descriptors (see https://docs.python.org/3/howto/descriptor.html)

In meta and base implementations, we use property data descriptors, so the following class:

```python
class Person(Base):
    name = fields.String(default="ahmed")
```

Should have different behavior when accessing `name` from a class or an objects, so, it will be converted by meta class to a class like:

```
class Person(Base):
    def __init__(self):
        self.__name = "ahmed"

    @property
    def get_name(self):
        return self.__name

    @property
    def set_name(self, value):
        self.__name == value

    name = property(get_name, set_name)
```

And accessing `name` from class and object levels will yield:

```python
Person.name  #=> <property object at 0x7efd89a259f8>


p = Person()
p.name  #=> "ahmed"
```


Parent relationship is supported too, every instance can have a parent object (which must be a `Base` type too)
"""
from types import SimpleNamespace

from jumpscale.core import events

from . import fields
from .factory import Factory, StoredFactory, DuplicateError
from .events import AttributeUpdateEvent


def get_field_property(name: str, field: fields.Field) -> property:
    """
    get a new property descriptor object for a field,
    this property will be used to enable getting/setting the actual value

    as the field only describes the type and other validation/conversion options,
    but do not hold the value itself, the vale will be held in the base instance

    the getter and setter will be called when an object is already created,
    and any field is accessed:

    ```python
    car = Car()
    print(car.color)  #=> getter will be called
    car.color = "red"  #=> setter will be called
    ```

    Args:
        name (str): field name
        field (fields.Field): field instance

    Returns:
        property: property descriptor (object)
    """

    def getter(self):
        """
        getter method this property

        will call `_get_value`, which would if the value is already defined
        and will get the default value if not

        Returns:
            any: the field value
        """
        return self._get_value(name, field)

    def setter(self, value):
        """
        a setter method for this property

        will call _set_value, which would do some checks:

        - validation: using field.validate_with_name
        - setting an attribute with inner_name in the base instance

        if it's set correctly, we will:

        - call `_attr_updated` of `self` with the name of this `field`
        - call `on_update` of the `field` with `self`

        Args:
            value (any): a value to be set for this field

        Raises:
            fields.ValidationError: in case the value is not valid
        """
        self._set_value(name, field, value)

        # call _attr_updated and on_update handlers
        self._attr_updated(name, value)
        if field.trigger_updates:
            field.on_update(self, value)

    return property(fget=getter, fset=setter)


class BaseMeta(type):
    """
    this class is used to get a new class with all field attributes replaced by property data descriptors.

    this should be used as a metaclass, example:

    ```python
    class ExampleWithFields(metaclass=BaseMeta):
        name = fields.String()
    ```
    """

    def __new__(cls, name: str, based: tuple, attrs: dict) -> type:
        """
        get a new class with all field attributes replaced by property data descriptors.

        Args:
            name (str): class name
            based (tuple): super class types (classes)
            attrs (dict): current attributes

        Returns:
            type: a new class
        """
        # will collect class fields
        # and also all fields of super classes, as we already have them in based
        cls_fields = {}
        super_fields = {}

        for super_cls in based:
            if hasattr(super_cls, "_fields"):
                super_fields.update(super_cls._fields)

        # update current attrs with super class fields
        attrs.update(super_fields)

        # now we maintain old attributes, but convert any attribute
        # with fields.Field type to property descriptor (property object)
        # using get_field_property
        new_attrs = {}
        for key in attrs:
            obj = attrs[key]
            if isinstance(obj, fields.Field):
                cls_fields[key] = obj
                new_attrs[key] = get_field_property(key, obj)
            else:
                # keep other attrs
                new_attrs[key] = obj

        new_class = super(BaseMeta, cls).__new__(cls, name, based, new_attrs)
        # set _fields attributes to cls_fields dict, so, we still have access to field objects
        new_class._fields = cls_fields
        return new_class


class Base(SimpleNamespace, metaclass=BaseMeta):
    def __init__(self, parent_=None, instance_name_=None, **values):
        """
        base class implementation for any class with fields which supports getting/setting raw data for any instance fields.

        any instance can have an optional name and a parent.

        ```python
        class Person(Base):
            name = fields.String()
            age = fields.Float()

        p = Person(name="ahmed", age="19")
        print(p.name, p.age)
        ```

        Args:
            parent_ (Base, optional): parent instance. Defaults to None.
            instance_name_ (str, optional): instance name. Defaults to None.
            **values: any given field values to initiate the instance with
        """
        self.__parent = parent_
        self.__instance_name = instance_name_

        self._factories = {}

        # now we create factories
        for name, field in self._get_fields().items():
            if isinstance(field, fields.Factory):
                value = field.factory_type(field.type, name_=name, parent_instance_=self)
                self._factories[name] = value
                setattr(self, f"__{name}", value)
                # if provided in values, remove it, as it's not needed
                if name in values:
                    values.pop(name)

        # setting other values
        self._set_data(values)

    def _get_fields(self):
        """
        get current defined field objects

        Returns:
            dict: fields dict as {name: field object}
        """
        return self._fields

    def _get_computed_fields(self):
        """
        get current defined field objects with compute function

        Returns:
            dict: fields dict as {name: field object}
        """
        return {name: field for name, field in self._fields.items() if field.computed}

    def _get_factories(self):
        """
        get sub-factory objects, which are defined by `fields.Factory`

        Returns:
            dict: factories as {name: factory object}
        """
        return self._factories

    def _get_embedded_objects(self):
        """
        get a list of embedded objects which are defined by `fields.Object`

        Returns:
            list: list of `Base` objects
        """
        return [getattr(self, name) for name, field in self._get_fields().items() if isinstance(field, fields.Object)]

    def _get_value(self, name, field):
        """
        get a field value

        Args:
            name (str): field name
            field (fields.Field): field object

        Returns:
            any: field value
        """
        # if computed, return the computed value
        if field.computed:
            return field.compute(self)

        # if it's already defined, just return it
        # we don't use hasattr here, because it uses getattr inside
        # it causes an infinite recursion here if the attr is not found
        # and also when __getattr__ is overridden
        inner_name = f"__{name}"
        if inner_name in self.__dict__:
            return getattr(self, inner_name)

        # if default is callable, get it
        if callable(field.default):
            default = field.default()
        else:
            default = field.default

        # use the actual name (not inner_name) to do validation and conversion...etc
        self._set_value(name, field, default)
        return self._get_value(name, field)

    def _set_value(self, name, field, value):
        """
        set a field value

        Args:
            name (str): field name
            field (fields.Field): field object
            value (any): value

        Raises:
            fields.ValidationError: raised if the value is not valid
        """
        if field.readonly:
            raise fields.ValidationError(f"'{name}' is a read only attribute")

        # accept if this is a raw value too
        value = field.from_raw(value)

        # validate
        field.validate_with_name(value, name)

        # set current instance as parent for embedded objects/instances
        if isinstance(field, fields.Object) and value:
            value._set_parent(self)

        # set as an internal attribute
        inner_name = f"__{name}"
        setattr(self, inner_name, value)

    def _get_data(self):
        """
        get a serializable dict from all values of all fields (except factories)

        ```python
        class Person(Base):
            name = fields.String()
            age = fields.Float()

        p = Person(name="ahmed", age=1.4)
        p._get_data()  #=> {'name': 'ahmed', 'age': '19'}
        p.to_dict() #=> {'name': 'ahmed', 'age': '19'}
        ```

        Returns:
            dict: data as dict with {name: value}
        """
        data = {}

        for name, field in self._get_fields().items():
            if isinstance(field, fields.Factory):
                # skip for factories for now
                continue
            if not field.stored:
                # skip non-stored fields too
                continue

            value = self._get_value(name, field)
            raw_value = field.to_raw(value)
            if isinstance(field, fields.Secret):
                data[f"__{name}"] = raw_value
            else:
                data[name] = raw_value

        return data

    def _set_data(self, new_data):
        """
        set values from dict to all fields (except factories)

        Args:
            new_data (dict): field values mapping
        """
        all_fields = self._get_fields()

        for name, value in new_data.items():
            if name in all_fields:
                try:
                    self._set_value(name, all_fields[name], value)
                except (fields.ValidationError, ValueError):
                    # should at least log validation and value errors
                    # this can happen in case of e.g. fields type change
                    pass

    def _attr_updated(self, name, value):
        """
        called when an attribute value is updated

        Args:
            name (str): attribute/field name
            value (any): value
        """
        event = AttributeUpdateEvent(self, name, value)
        events.notify(event)

    def validate(self):
        """
        validate all fields of current instance
        """
        for name, field in self._get_fields().items():
            field.validate_with_name(getattr(self, name), name)

    @property
    def parent(self):
        return self.__parent

    def _set_parent(self, parent):
        """
        set current parent instance

        Args:
            parent (Base): base object/instance
        """
        self.__parent = parent

    @property
    def instance_name(self):
        return self.__instance_name

    def _set_instance_name(self, name):
        """
        set current instance name

        Args:
            name (str): name
        """
        self.__instance_name = name

    to_dict = _get_data

    @classmethod
    def from_dict(cls, data):
        """
        get an instance from a dict

        ```python
        class Person(Base):
            name = fields.String()
            age = fields.Float()

        p = Person.from_dict({"name": "ahmed", "age": 19})
        print(p.name, p.age)  #=> ahmed 19
        ```

        Args:
            data (dict): values dict

        Returns:
            Base: an instance from current `Base` type
        """
        return cls(**data)

    def __eq__(self, other):
        """
        compare self to `other`, which must be of the same type.

        this just compares the data of two objects.

        Args:
            other (Base): other object of the same type of `self`

        Returns:
            bool: `True` if equal, `False` otherwise
        """
        return type(self) == type(other) and self.to_dict() == other.to_dict()



"""This module execute process on system and manage them
for example
```
#to create a process
rc, out, err = j.sals.process.execute("ls", cwd="/tmp", showout=True)
#this executes ls command on dir "/tmp" showing output from stdout
#rc -> contains exit status
#out -> the actual output
#err -> in case an error happened this var will contains the error msg

j.sals.process.is_active(10022)
#checks if a process with this pid is active or not

j.sals.process.kill(10022, sig=signal.SIGTERM.value)
#kill a process with pid 10022 with SIGTERM

j.sals.process.get_pid_by_port(8000)
#gets pid of the process listenning on port 8000
```
"""


import math
import os
import os.path
import random
import re
import select
import signal
import subprocess
import sys
import time
from collections import defaultdict
from subprocess import Popen

import psutil
import socket
from jumpscale.loader import j


def execute(
    cmd,
    showout=False,
    cwd=None,
    shell="/bin/bash",
    timeout=600,
    asynchronous=False,
    env=None,
    replace_env=False,
    die=False,
):
    """
    execute a command.

    accepts command as a list too, with auto-escaping.

    Arguments:
        cmd (str or list): command to be executed, e.g. `"ls -la"` or `["ls", "-la"]

    Keyword Arguments:
        showout (bool): show stdout of the command (default: False)
        cwd (str): specify a working directory for the command (default: None)
        shell (str): specify a shell to execute the command (default: "/bin/bash")
        timeout (int): timeout before kill the process (default: 600)
        asynchronous (bool): execute in asynchronous mode or not (default: False)
        env (dict): add environment variables here (default: {})
        replace_env (bool): replace entire environment with env (default: False)
        die (bool): die if command failed (default: False)

    Returns:
        tuple: (rc, out, err)
    """
    return j.core.executors.run_local(
        cmd=cmd,
        hide=not showout,
        cwd=cwd,
        shell=shell,
        timeout=timeout,
        asynchronous=asynchronous,
        env=env or {},
        replace_env=replace_env,
        warn=not die,
    )


def is_alive(pid):
    """Checks if pid is Running

    Arguments:
        pid (int) -- pid of the process to be checked

    Returns:
        [bool] -- True if process is running
    """
    pid = int(pid)
    return psutil.pid_exists(pid)


def is_installed(cmd):
    """[summary]
    Checks if a specific command is available on system e.g. curl
    Arguments:
        cmd {str} -- command to be checked

    Returns:
        [bool] -- True if command is installed
    """
    rc, _, _ = execute("which %s" % cmd, die=False)
    if rc:
        return False
    else:
        return True


def kill(pid, sig=signal.SIGTERM.value):
    """Kill a process with a signal

    Arguments:
        pid (int) -- pid of the process to be killed

    Keyword Arguments:
        sig {int]} -- which signal you want to kill the process with (default: {signal.SIGTERM.value})

    Raises:
        j.exceptions.RuntimeError: in case killing process failed

    Returns:
        [type] -- [description]
    """
    pid = int(pid)
    sig = int(sig)
    proc = psutil.Process(pid)
    try:
        proc.send_signal(sig)
        return True
    except Exception as e:
        raise j.exceptions.RuntimeError("Could not kill process with id %s.\n%s" % (pid, e))


def ps_find(name):
    """find process by name

    Arguments:
        name {str} -- process name

    Returns:
        [bool] -- True if process is found
    """
    for proc in psutil.process_iter():
        if proc.name() == name:
            return True
    return False


def kill_all(name, sig=signal.SIGKILL):
    """Kill all processes with a given name

    Arguments:
        name {str} -- process name

    Keyword Arguments:
        sig (int) -- signal number (default: {signal.SIGKILL})
    """
    sig = int(sig)
    for proc in psutil.process_iter():
        if proc.name() == name:
            kill(proc.pid, sig)


def get_pids_filtered_sorted(filterstr, sortkey=None):
    """Get pids of process by a filter string and optionally sort by sortkey

    Arguments:
        filterstr {[str]} -- filter string.

    Keyword Arguments:
        sortkey {[str]} -- sort key for ps command (default: {None})
        sortkey can be one of the following:
        %cpu           cpu utilization of the process in
        %mem           ratio of the process's resident set size  to the physical memory on the machine, expressed as a percentage.
        cputime        cumulative CPU time, "[DD-]hh:mm:ss" format.  (alias time).
        egid           effective group ID number of the process as a decimal integer.  (alias gid).
        egroup         effective group ID of the process.  This will be the textual group ID, if it can be obtained and the field width permits, or a decimal representation otherwise.  (alias group).
        euid           effective user ID (alias uid).
        euser          effective user name.
        gid            see egid.  (alias egid).
        pid            a number representing the process ID (alias tgid).
        ppid           parent process ID.
        psr            processor that process is currently assigned to.
        start_time     starting time or date of the process.


    Returns:
        [list(int)] -- processes pids
    """
    if sortkey is not None:
        cmd = "ps aux --sort={sortkey} | grep '{filterstr}'".format(filterstr=filterstr, sortkey=sortkey)
    else:
        cmd = "ps ax | grep '{filterstr}'".format(filterstr=filterstr)
    rc, out, err = execute(cmd)
    # print out
    found = []
    for line in out.split("\n"):
        if line.find("grep") != -1 or line.strip() == "":
            continue
        if line.strip() != "":
            if line.find(filterstr) != -1:
                line = line.strip()
                if sortkey is not None:
                    found.append(int([x for x in line.split(" ") if x][1]))
                else:
                    found.append(int(line.split(" ")[0]))
    return found


def get_filtered_pids(filterstr, excludes=None):
    """Get pids filtered by filterstr and execludes

    Arguments:
        filterstr {str} -- filter string.

    Keyword Arguments:
        excludes {list(str)} -- execlude list (default: {None})

    Returns:
        [list(int)] -- pids
    """
    excludes = excludes or []
    cmd = "ps ax | grep '%s'" % filterstr
    rc, out, err = j.core.executors.run_local(cmd)
    # print out
    found = []

    def checkexclude(c, excludes):
        for item in excludes:
            c = c.lower()
            if c.find(item.lower()) != -1:
                return True
        return False

    for line in out.split("\n"):
        if line.find("grep") != -1 or line.strip() == "":
            continue
        if line.strip() != "":
            if line.find(filterstr) != -1:
                line = line.strip()
                if not checkexclude(line, excludes):
                    # print "found pidline:%s"%line
                    found.append(int(line.split(" ")[0]))
    return found


def get_pids_filtered_by_regex(regex_list, excludes=None):
    """get pids of a process filtered by Regex list

    Arguments:
        regex_list {list(str)} -- list of regex expressions

    Keyword Arguments:
        excludes {list(str)} -- list of excludes (default: {None})

    Returns:
        [list(int)] -- list of pids
    """
    excludes = excludes or []
    res = []
    for process in psutil.process_iter():
        try:
            cmdline = process.cmdline()
        except psutil.NoSuchProcess:
            cmdline = None
        except psutil.AccessDenied:
            cmdline = None
        if cmdline:
            name = " ".join(cmdline)
            for r in regex_list:
                if name.strip() != "" and re.match(r, name):
                    res.append(process.pid)
    return res


def check_start(cmd, filterstr, nrinstances=1, retry=1):
    """Run command and check if it is started based on filterstr

    Arguments:
        cmd {str} -- command to be executed
        filterstr {str} -- filter string

    Keyword Arguments:
        instances (int) -- number of needed instances (default: {1})
        retry (int) -- number of retries (default: {1})

    Raises:
        j.exceptions.RuntimeError: will be raised if we didn't reach number of required instances
    """
    found = get_filtered_pids(filterstr)
    for i in range(retry):
        if len(found) == nrinstances:
            return
        # print "START:%s"%cmd
        execute(cmd)
        time.sleep(1)
    found = get_filtered_pids(filterstr)
    if len(found) != nrinstances:
        raise j.exceptions.RuntimeError(
            "could not start %s, found %s nr of instances. Needed %s." % (cmd, len(found), nrinstances)
        )


def check_stop(cmd, filterstr, retry=1, nrinstances=0):
    """Executes a stop command and check if it is already stopped based on filterstr

    Arguments:
        cmd {str} -- command to be executed
        filterstr {str} -- filter string

    Keyword Arguments:
        retry (int) -- number of retries (default: {1})
        nrinstances (int) -- number of instances after stop (default: {0})

    Raises:
        j.exceptions.RuntimeError: if nr of instances not matched
    """

    found = get_filtered_pids(filterstr)
    for i in range(retry):
        if len(found) == nrinstances:
            return
        # print "START:%s"%cmd
        execute(cmd, die=False)
        time.sleep(1)
        found = get_filtered_pids(filterstr)
        for item in found:
            kill(int(item), 9)
        found = get_filtered_pids(filterstr)

    if len(found) != 0:
        raise j.exceptions.RuntimeError("could not stop %s, found %s nr of instances." % (cmd, len(found)))


def get_pids(process_name, match_predicate=None):
    """Get process ID(s) for a given process

    Arguments:
        process {str} -- process name

    Keyword Arguments:
        match_predicate {callable} -- function that does matching between
        found processes and the targested process, the function should accept
        two arguments and return a boolean, defaults to None (default: {None})

    Raises:
        j.exceptions.RuntimeError: [description]
        j.exceptions.NotImplemented: [description]

    Returns:
        [list(int)] -- list of pids
    """
    # default match predicate
    # why aren't we using psutil ??
    def default_predicate(target, given):
        return target.strip().lower() in given.lower()

    if match_predicate is None:
        match_predicate = default_predicate

    if process_name is None:
        raise j.exceptions.RuntimeError("process cannot be None")
    if j.data.platform.is_unix():
        pids = set()
        for process in get_processes():
            try:
                pid = process.pid
                if not isinstance(pid, int):
                    continue
                name = process.name()
                if match_predicate(process_name, name):
                    pids.add(pid)
                elif match_predicate(process_name, process.exe()):
                    pids.add(pid)
                else:
                    cmdline = process.cmdline()
                    if cmdline and cmdline[0]:
                        if match_predicate(process_name, cmdline[0]):
                            pids.add(pid)
            except (psutil.Error, FileNotFoundError):
                continue
        return list(pids)
    else:
        raise j.exceptions.NotImplemented("getProcessPid is only implemented for unix")


def get_my_process():
    """get process object of current process

    Returns:
        [psutil.Process] -- process object
    """
    return get_process_object(os.getpid())


def get_process_object(pid, die=True):
    """Get Process object of a process id

    Arguments:
        pid (int) -- pid of the process

    Keyword Arguments:
        die {bool} -- die if process not found (default: {True})

    Raises:
        psutil.NoSuchProcess: if process not found and die = True

    Returns:
        [psutil.Process] -- process object
    """
    try:
        return psutil.Process(pid)
    except psutil.NoSuchProcess as e:
        if die:
            raise e
        else:
            return None


def get_user_processes(user):
    """Get all process for a specific user

    Arguments:
        user {str} -- username

    Returns:
        [list(int)] -- list of process pids for that user
    """
    result = []
    for process in psutil.process_iter():
        if process.username() == user:
            result.append(process.pid)
    return result


def kill_user_processes(user):
    """Kill all processes for a specific user

    Arguments:
        user {str} -- username
    """
    for pid in get_user_processes(user):
        kill(pid)


def get_similar_processes():
    """Gets similar processes to current process

    Returns:
        [list(psutil.Process)] -- list of similar process
    """
    myprocess = get_my_process()
    result = []
    for item in psutil.process_iter():
        try:
            if item.cmdline() == myprocess.cmdline():
                result.append(item)
        except psutil.NoSuchProcess:
            pass
    return result


def check_running(process, min=1):
    """Checks if a process is running

    Arguments:
        process {str} -- process name to be checked

    Keyword Arguments:
        min (int) -- min number of instances required to be running (default: {1})

    Returns:
        [bool] -- true if process is running
    """
    if j.data.platform.is_linux():
        pids = get_pids(process)
        if len(pids) >= min:
            return True
        return False


def check_process_for_pid(pid, process_name):
    """Check whether a given pid actually does belong to a given process name.

    Arguments:
        pid (int) -- process pid
        process {str} -- process name

    Returns:
        [bool] -- True if process_name matched process name of the pid
    """
    pid = int(pid)
    proc = psutil.Process(pid)
    return proc.name() == process_name


def set_env_var(varnames, varvalues):
    """Set the value of the environment variables C{varnames}. Existing variable are overwritten

    Arguments:
        varnames {list(str)} --  A list of the names of all the environment variables to set
        varvalues {list(str)} -- A list of all values for the environment variables

    """
    try:
        for i in range(len(varnames)):
            os.environ[varnames[i]] = str(varvalues[i]).strip()
    except Exception as e:
        raise j.exceptions.RuntimeError(e)


def get_pid_by_port(port):
    """Returns pids of the process that is listening on the given port

    Arguments:
        port (int) -- port number

    Returns:
        int -- pid of process that listen on that port
    """

    process = get_process_by_port(port)
    if process is None:
        return []
    return process.pid


def kill_process_by_name(name, sig=signal.SIGTERM.value, match_predicate=None):
    """Kill all processes for a given command

    Arguments:
        name {str} -- Name of the command that started the process(s)

    Keyword Arguments:
        sig {bool} -- os signal to send to the process(s) (default: {signal.SIGTERM.value})
        match_predicate {callable} -- function that does matching between
            found processes and the targested process, the function should accept
            two arguments and return a boolean (default: {None})
    """

    pids = get_pids(name, match_predicate=match_predicate)
    for pid in pids:
        kill(pid, sig)


def kill_process_by_port(port):
    """Kill process by port

    Arguments:
        port (int) -- port number
    """
    port = int(port)
    for pid in get_pid_by_port(port):
        kill(pid)


def is_port_listening(port):
    """check if the port is being used by any process

    Args:
        port (int): port number

    Returns:
        Bool: True if port is used else False
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        result = s.connect_ex(('127.0.0.1', port))
    return result == 0


def get_process_by_port(port):
    """Returns the full name of the process that is listening on the given port

    Arguments:
        port (int) -- the port for which to find the command

    Raises:
        Runtime Error if the process is not accessible by the user

    Returns:
        [psutil.Process] -- process object
        None -- No process found
    """
    pcons = [proc for proc in psutil.net_connections() if proc.laddr.port == port and proc.status == "LISTEN"]
    if pcons:
        pid = pcons[0].pid
        if not pid:
            raise j.exceptions.Runtime("No pid found maybe permission denied on the process")
        return psutil.Process(pid)


def get_defunct_processes():
    """Gets defunc processes

    Returns:
        [list(int)] -- list of processes pids
    """
    _, out, _ = execute("ps ax")
    llist = []
    for line in out.split("\n"):
        if line.strip() == "":
            continue
        if line.find("<defunct>") != -1:
            # print "defunct:%s"%line
            line = line.strip()
            pid = line.split(" ", 1)[0]
            pid = int(pid.strip())
            llist.append(pid)

    return llist


def get_processes():
    """
    get an interator for all running processes

    Yields:
        generator: for all processes
    """
    yield from psutil.process_iter()


def get_processes_info():
    """
    Get information for top 25 running processes sorted by memory usage

    Returns:
        [list(dict)] -- list of processes info
    """
    processes_list = []
    for proc in get_processes():
        try:
            # Fetch process details as dict
            pinfo = proc.as_dict(attrs=["pid", "name", "username"])
            pinfo["rss"] = proc.memory_info().rss / (1024 * 1024)
            pinfo["ports"] = []
            try:
                connections = proc.connections()
            except psutil.Error:
                continue
            if connections:
                for conn in connections:
                    pinfo["ports"].append({"port": conn.laddr.port, "status": conn.status})
            # Append dict to list
            processes_list.append(pinfo)
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    processes_list = sorted(processes_list, key=lambda procObj: procObj["rss"], reverse=True)
    return processes_list[:25]


def get_ports_mapping(status=psutil.CONN_LISTEN):
    """
    get a mapping for process to ports with a status filter

    it will skip any process in case of errors (e.g. permission error)

    example:

    ```python
    j.sals.process.get_ports_mapping(psutil.CONN_ESTABLISHED)
    ```

    or

    ```
    j.sals.process.get_ports_mapping("ESTABLISHED")
    ```

    Args:
        status (psutil.CONN_CONSTANT): `psutil` CONN_* constant as a filter. Defaults to psutil.CONN_LISTEN.

    Returns:
        defaultdict: a mapping between process and ports
    """
    ports = defaultdict(list)

    for process in get_processes():
        try:
            connections = process.connections()
        except psutil.Error:
            continue

        if connections:
            for conn in connections:
                if conn.status == status:
                    ports[process].append(conn.laddr.port)

    return ports


def get_memory_usage():
    """
    Get memory status

    Returns:
        dict -- memory status info
    """
    memory_usage = {}
    memory_data = dict(psutil.virtual_memory()._asdict())
    memory_usage["total"] = math.ceil(memory_data.get("total") / (1024 * 1024 * 1024))
    memory_usage["used"] = math.ceil(memory_data.get("used") / (1024 * 1024 * 1024))
    memory_usage["percent"] = memory_data.get("percent")
    return memory_usage


def get_environ(pid):
    """Gets env vars for a specific process based on pid

    Arguments:
        pid (int) -- process pid

    Returns:
        [dict] -- dict of env variables
    """
    pid = int(pid)
    return psutil.Process(pid).environ()


def in_docker():
    """will check if we are in a docker

    Returns:
        Bool: True if in docker - False if not
    """
    rc, out, _ = j.sals.process.execute("cat /proc/1/cgroup", die=False, showout=False)
    if rc == 0 and "/docker/" in out:
        return True
    return False


def in_host():
    """will check if we are in a host

    Returns:
        Bool: True if in host - False if not
    """
    return not in_docker()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jumpscale',
 'jumpscale.clients',
 'jumpscale.clients.docker',
 'jumpscale.clients.gedis',
 'jumpscale.clients.git',
 'jumpscale.clients.redis',
 'jumpscale.clients.sshclient',
 'jumpscale.clients.sshkey',
 'jumpscale.clients.zdb',
 'jumpscale.core.application',
 'jumpscale.core.base',
 'jumpscale.core.base.store',
 'jumpscale.core.config',
 'jumpscale.core.db',
 'jumpscale.core.dirs',
 'jumpscale.core.events',
 'jumpscale.core.exceptions',
 'jumpscale.core.executors',
 'jumpscale.core.logging',
 'jumpscale.data.bcdb',
 'jumpscale.data.bcdb.models',
 'jumpscale.data.cache',
 'jumpscale.data.encryption',
 'jumpscale.data.fake',
 'jumpscale.data.hash',
 'jumpscale.data.idgenerator',
 'jumpscale.data.inifile',
 'jumpscale.data.nacl',
 'jumpscale.data.platform',
 'jumpscale.data.random_names',
 'jumpscale.data.schema',
 'jumpscale.data.serializers',
 'jumpscale.data.tarfile',
 'jumpscale.data.terminaltable',
 'jumpscale.data.text',
 'jumpscale.data.time',
 'jumpscale.data.types',
 'jumpscale.entry_points',
 'jumpscale.sals.fs',
 'jumpscale.sals.hostsfile',
 'jumpscale.sals.nettools',
 'jumpscale.sals.process',
 'jumpscale.sals.unix',
 'jumpscale.servers.gedis',
 'jumpscale.servers.gedis_http',
 'jumpscale.servers.openresty',
 'jumpscale.servers.rack',
 'jumpscale.shell',
 'jumpscale.threesdk',
 'jumpscale.tools.alerthandler',
 'jumpscale.tools.codeloader',
 'jumpscale.tools.console',
 'jumpscale.tools.depsresolver',
 'jumpscale.tools.errorhandler',
 'jumpscale.tools.git',
 'jumpscale.tools.highlighter',
 'jumpscale.tools.http',
 'jumpscale.tools.jinja2',
 'jumpscale.tools.keygen',
 'jumpscale.tools.schemac',
 'jumpscale.tools.schemac.plugins',
 'jumpscale.tools.startupcmd',
 'jumpscale.tools.syncer',
 'jumpscale.tools.timer']

package_data = \
{'': ['*'],
 'jumpscale': ['install/*'],
 'jumpscale.servers.openresty': ['templates/*']}

install_requires = \
['GitPython>=3.0,<4.0',
 'PyYAML>=5.1,<6.0',
 'Whoosh>=2.7.4,<3.0.0',
 'arrow>=0.15.7,<0.16.0',
 'base58>=2.0.0,<3.0.0',
 'better-exceptions>=0.2.2,<0.3.0',
 'bottle>=0.12.18,<0.13.0',
 'captcha>=0.3,<0.4',
 'certifi>=2019.6,<2020.0',
 'click>=7.0,<8.0',
 'colorama>=0.4.1,<0.5.0',
 'cryptocompare>=0.6.4,<0.7.0',
 'dill>=0.3.0,<0.4.0',
 'distro>=1.4,<2.0',
 'docker>=4.2.0,<5.0.0',
 'fabric>=2.4,<3.0',
 'faker>=2.0,<3.0',
 'gevent==1.4.0',
 'jinja2>=2.11.1,<3.0.0',
 'libtmux>=0.8.2,<0.9.0',
 'loguru>=0.3.2,<0.4.0',
 'msgpack>=0.6.1,<0.7.0',
 'netaddr>=0.7.19,<0.8.0',
 'pdoc3>=0.6.3,<0.7.0',
 'pillow>=6.2.2,<7.0.0',
 'prompt-toolkit<3.0.0',
 'psutil>=5.7.0,<6.0.0',
 'ptpython>=2.0,<3.0',
 'pudb>=2019.1,<2020.0',
 'pycparser>=2.20,<3.0',
 'pylzma>=0.5.0,<0.6.0',
 'pynacl==1.3.0',
 'pytoml>=0.1.21,<0.2.0',
 'redis>=3.3,<4.0',
 'secretconf>=0.1.2,<0.2.0',
 'terminaltables>=3.1,<4.0',
 'watchdog>=0.9.0,<0.10.0',
 'zipp>=3.1.0,<4.0.0']

entry_points = \
{'console_scripts': ['jsctl = jumpscale.entry_points.jsctl:cli',
                     'jsng = jumpscale.entry_points.jsng:run',
                     'jsync = jumpscale.entry_points.jsync:cli',
                     'usershell = jumpscale.entry_points.usershell:run']}

setup_kwargs = {
    'name': 'js-ng',
    'version': '11.0.0a1',
    'description': '',
    'long_description': None,
    'author': 'xmonader',
    'author_email': 'xmonader@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import msgpack


def dumps(obj):
    """dump dict object into msgpack stream 
    
    Arguments:
        obj (dict) : the dict which will be dumped     
    
    Returns:
        string : the msgpack stream
    """
    return msgpack.packb(obj, use_bin_type=True)


def loads(s):
    """loads the data from msgpack string into dict
    
    Arguments:
        s (string) : the msgpack stream
    
    Returns:
        dict : the loaded data from msgpack stram
    """
    if isinstance(s, (bytes, bytearray)):
        return msgpack.unpackb(s, raw=False)
    return False



"""This module gives you all the facilities to communicate with gedis server

Connecting to a gedis server
```
JS-NG> gedis = j.clients.gedis.get("local")
JS-NG> gedis.list_actors()
['system']
```

Registering actor
```
JS-NG> gedis.actors.system.register_actor("greeter", "/home/ahmed/wspace/threefoldtech/js-ng/jumpscale/servers/gedis/example_greeter.py")
1
```

Listing actors
```
JS-NG> gedis.list_actors()
['system', 'greeter']
```

Documentation of an actor

```
JS-NG> gedis.ppdoc("greeter")
{
  "add2": {
    "args": [
      "a",
      "b"
    ],
    "doc": "Add two args\n        \n        "
  },
  "hi": {
    "args": [],
    "doc": "returns hello world\n        "
  },
  "info": {
    "args": [
      "result",
      "members",
      "name",
      "attr"
    ],
    "doc": ""
  },
  "ping": {
    "args": [],
    "doc": "\n        \n        "
  }
}
```
Invoking an actor method
```
JS-NG> gedis.execute("greeter", "hi")
b'hello world'

JS-NG> gedis.execute("greeter", "ping")
b'pong no?'

JS-NG> gedis.execute("greeter", "add2", "first", "second")
b'firstsecond'
```
Invoking actor method with attribute access
```
JS-NG> gedis.actors.greeter.hi()
b'hello world'

JS-NG> gedis.actors.greeter.add2("a", "b")
b'ab'
```
"""


def export_module_as():

    from jumpscale.core.base import StoredFactory

    from .gedis import GedisClient

    return StoredFactory(GedisClient)



"""Module to help syncing multiple machines with specific directories you have.
used in the jsync tool.
```
JS-NG> xmonader = j.clients.sshkey.new("xmonader")
JS-NG> xmonader.private_key_path = "/home/xmonader/.ssh/id_rsa"
JS-NG> xmonader.load_from_file_system()
JS-NG> xmonader.save()
JS-NG> xmonader = j.clients.sshclient.new("xmonader")
JS-NG> xmonader.sshkey = "xmonader"
JS-NG> s = j.tools.syncer.Syncer(["xmonader"], {"/home/xmonader/wspace/tfchain-py":"/tmp/tfchain-py"})
JS-NG> s.start()
2019-09-03T11:38:47.183394+0200 - paths: {'/home/xmonader/wspace/tfchain-py': '/tmp/tfchain-py'}
```
"""

from jumpscale.loader import j
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer
import gevent
from typing import List, Dict, Optional

DEFAULT_IGNORED_PATTERNS = [".git", ".pyc", "__pycache__", ".swp", ".swx"]


class Syncer(PatternMatchingEventHandler):
    def __init__(
        self,
        sshclients_names: List[str],
        paths: Dict[str, str],
        patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
        ignore_directories: Optional[List[str]] = False,
        case_sensitive: bool = True,
    ):
        """Creates new syncer tool

        Arguments:
            sshclients_names {List[str]} -- list of sshclient names
            paths {Dict[str, str]} -- paths to watch src/dest form of dict {'/tmp/myproj':'/root/proj'}

        Keyword Arguments:
            patterns {Optional[List[str]]} -- optional list of patterns to watch (default: {None})
            ignore_patterns {Optional[List[str]]} -- patterns to ignore, e.g .git, __pycache__ (default: {None})
            ignore_directories {Optional[List[str]]} -- directories to ignore (default: {False})
            case_sensitive {bool} -- case sensitive watching  (default: {True})

        Returns:
            Syncer -- Syncer object
        """
        ignore_patterns = ignore_patterns or DEFAULT_IGNORED_PATTERNS
        super().__init__(patterns, ignore_patterns, ignore_directories, case_sensitive)
        self.observer = Observer()
        self.sshclients_names = sshclients_names
        self.paths = paths or {}  # src:dst

    def _get_dest_path(self, src_path: str) -> str:
        """returns destination path in remote machine

        Arguments:
            src_path {str} -- path in source machine

        Returns:
            str -- path in remote machine
        """
        j.logger.debug(f"paths: {self.paths} and path: {src_path}")

        for path in self.paths.keys():
            if path.startswith(src_path):
                return self.paths[src_path]

    def _rewrite_path_for_dest(self, src_path: str) -> str:
        """rewrite src_path to remote_path
        e.g
            local: /tmp/myproj/file.py
            remote: /root/myproj/file.py

        Arguments:
            src_path {str} -- source machine path

        Returns:
            str -- rewritten path for remote
        """
        src_path = str(src_path)
        for path in self.paths.keys():
            if src_path.startswith(path):
                return src_path.replace(path, self.paths[path])

    def _get_sshclients(self):
        """Returns list of sshclient objects.

        Returns:
            List[SSHClient] -- list of ssh clients
        """
        clients = []
        for name in self.sshclients_names:
            clients.append(j.clients.sshclient.get(name))
        return clients

    def sync(self):
        """Sync directory structure and files

        """
        j.logger.debug(f"paths: {self.paths}")

        def ensure_dirs():
            """For every directory in watched paths we make sure it's full path exists on remote.

            """
            for path in self.paths:
                for src_dir in j.sals.fs.walk_dirs(path):
                    dest_dir = str(self._rewrite_path_for_dest(src_dir))
                    for cl in self._get_sshclients():
                        j.logger.debug(f"making dir {dest_dir}")
                        cl.sshclient.run(f"mkdir -p {dest_dir}")
                        self.observer.schedule(self, src_dir)

        def sync_file(e):
            """Sync single file to all registered sshclients

            Arguments:
                e {str} -- file path
            """
            dest_path = self._rewrite_path_for_dest(e)
            j.logger.debug(f"syncing {e} to machines into {dest_path}")

            for cl in self._get_sshclients():
                cl.sshclient.run(f"mkdir -p {j.sals.fs.parent(dest_path)}")
                cl.sshclient.sftp.put(e, self._rewrite_path_for_dest(e))

        def filter_ignored(e):
            return True

        ensure_dirs()

        for path in self.paths:
            for f in j.sals.fs.walk_files(path, sync_file):
                sync_file(f)

    def start(self, sync=True):
        """Start syncing/watching paths to remote machines

        Keyword Arguments:
            sync {bool} -- sync dirs/files first (default: {True})
        """
        if sync:
            self.sync()

        for path in self.paths.keys():
            self.observer.schedule(self, path)

        self.observer.start()
        try:
            while True:
                gevent.sleep(0.1)
        except KeyboardInterrupt:
            self.observer.unschedule_all()
            self.observer.stop()
        self.observer.join()

    def on_moved(self, event):
        super().on_moved(event)

        what = "directory" if event.is_directory else "file"
        j.logger.info(f"Moved {what}: from {event.src_path} to {event.dest_path}")
        dest_path = self._rewrite_path_for_dest(event.dest_path)
        j.logger.debug(f"will move to {dest_path}")
        j.logger.debug(f"will delete original in {self._rewrite_path_for_dest(event.src_path)}")
        for cl in self._get_sshclients():
            if not event.is_directory:
                try:
                    # in case file is moved
                    cl.sshclient.sftp.put(event.dest_path, dest_path)
                    cl.sshclient.run(f"rm {self._rewrite_path_for_dest(event.src_path)}")
                except:
                    j.logger.debug(f"Ignoring {dest_path}. Path was not found during move event")
            else:
                # in case file is directory
                cl.sshclient.sftp.posix_rename(self._rewrite_path_for_dest(event.src_path), dest_path)

    def on_created(self, event):
        super().on_created(event)
        what = "directory" if event.is_directory else "file"
        j.logger.debug(f"Created {what}: {event.src_path}")

        dest_path = self._rewrite_path_for_dest(event.src_path)
        j.logger.debug(f"will create in {dest_path}")

        for cl in self._get_sshclients():
            if what == "directory":
                cl.sshclient.run(f"mkdir -p {dest_path}")
                self.observer.schedule(self, event.src_path)
            else:
                try:
                    cl.sshclient.run(f"mkdir -p {j.sals.fs.parent(dest_path)}")
                    cl.sshclient.run(f"touch {dest_path}")
                except:
                    j.logger.debug(f"Ignoring {dest_path}. Path was not found during create event")

    def on_deleted(self, event):
        super().on_deleted(event)

        what = "directory" if event.is_directory else "file"
        j.logger.debug(f"Deleted {what}: {event.src_path}")

        dest_path = self._rewrite_path_for_dest(event.src_path)
        j.logger.debug(f"will delete in {dest_path}")
        for cl in self._get_sshclients():
            if what == "directory":
                cl.sshclient.run(f"rm -rf {dest_path}")
            else:
                try:
                    cl.sshclient.run(f"rm {dest_path}")
                except:
                    j.logger.debug(f"Ignoring {dest_path}. Path was not found during delete event")

    def on_modified(self, event):
        super().on_modified(event)
        what = "directory" if event.is_directory else "file"
        j.logger.debug(f"Modified {what}: {event.src_path}")

        dest_path = self._rewrite_path_for_dest(event.src_path)
        j.logger.debug(f"will modify in {dest_path}")

        for cl in self._get_sshclients():
            if what == "directory":
                j.logger.debug(f"Folder {dest_path} was modified")
            else:
                try:
                    cl.sshclient.sftp.put(event.src_path, dest_path)
                except:
                    j.logger.debug(f"Ignoring {dest_path}. Path was not found during modify event")



'''
# schemac 

Schemac is a tool used to convert (transpile) the schemas defined in jsx systems into the new objects definitions in js-ng

## Example

In this example there're bunch of types (bools, int, string, dict, time, date, enumerations, objects, list of objects, email) defined in jsx old style.

### jsx schema

```python
schema = """
@url = despiegk.test
listany = (LO)
llist2 = "" (LS) #L means = list, S=String
llist3     = [1,2,3] (LF)
today = (D)
now = (T)
info = (dict)
theemail = (email)
status = "on,off" (E) 
happy = "yes, no" (E)
&nr    = 4
obj = (O)!hamada.test
lobjs = (LO) !hamada.test

date_start = 0 (I)
description* = "hello world"
description2 ** = 'a string' (S)
llist4*** = [1,2,3] (LI)
llist5 = [1,2,3] (LI)
llist6 = [1,2,3] (LI)
U = 0.0
nrdefault = 0
nrdefault2 = (I)
nrdefault3 = 0 (I)

@url = hamada.test
a = (I)
name = (S)
mood = "stressed,sleeping" (E)
"""
```

### Converting to the new system

We expect that to expand or convert to plain old python classes, with dependency resolution.



```python
    c = j.tools.schemac.get_compiler(schema, "python")
    assert c

    assert c._schema_text
    assert c.lang == "python"
    assert c.generator
    parsed_schemas = c.parse()  # parse schema now
    generated_python =c.generator.generate(parsed_schemas)
    print(generated_python)
```

### Generated file

```python
#GENERATED CLASS DONT EDIT
from jumpscale.core.base import Base, fields
from enum import Enum

class Status(Enum):
    On = 0
    Off = 1

class Happy(Enum):
    Yes = 0
    No = 1

class Mood(Enum):
    Stressed = 0
    Sleeping = 1



class HamadaTest(Base):
    a = fields.Integer()
    name = fields.String(default="")
    mood = fields.Enum(Mood)



class DespiegkTest(Base):
    listany = fields.List(fields.Object(Base))
    llist2 = fields.List(fields.String())
    llist3 = fields.List(fields.Float())
    today = fields.DateTime()
    now = fields.Time()
    info = fields.Typed(dict)
    theemail = fields.Email()
    status = fields.Enum(Status)
    happy = fields.Enum(Happy)
    nr = fields.String(default="4")
    obj = fields.Object(HamadaTest)
    lobjs = fields.List(fields.Object(HamadaTest))
    date_start = fields.Integer(default=0)
    description = fields.String(default="hello world")
    description2 = fields.String(default="a string")
    llist4 = fields.List(fields.Integer())
    llist5 = fields.List(fields.Integer())
    llist6 = fields.List(fields.Integer())
    U = fields.String(default="0.0")
    nrdefault = fields.String(default="0")
    nrdefault2 = fields.Integer()
    nrdefault3 = fields.Integer(default=0)

```
'''

def get_compiler(schema_text, lang="python"):
    from .compiler import Compiler

    return Compiler(lang, schema_text)



"""This module helps genearting random names, mainly for container names.

```
JS-NG> j.data.random_names.random_name()                                                            
'loving_borg'

JS-NG> j.data.random_names.random_name()                                                            
'quirky_wiles'
```

"""

from random import choice

left_names = [
    "albattani",
    "allen",
    "almeida",
    "antonelli",
    "agnesi",
    "archimedes",
    "ardinghelli",
    "aryabhata",
    "austin",
    "babbage",
    "banach",
    "banzai",
    "bardeen",
    "bartik",
    "bassi",
    "beaver",
    "bell",
    "benz",
    "bhabha",
    "bhaskara",
    "black",
    "blackburn",
    "blackwell",
    "bohr",
    "booth",
    "borg",
    "bose",
    "boyd",
    "brahmagupta",
    "brattain",
    "brown",
    "burnell",
    "buck",
    "burnell",
    "cannon",
    "carson",
    "cartwright",
    "chandrasekhar",
    "chaplygin",
    "chatelet",
    "chatterjee",
    "chebyshev",
    "cocks",
    "cohen",
    "chaum",
    "clarke",
    "colden",
    "cori",
    "cray",
    "wiles",
    "williams",
    "williamson",
    "wilson",
    "wing",
    "wozniak",
    "wright",
    "wu",
    "yalow",
    "yonath",
    "zhukovsky",
]
right_names = [
    "admiring",
    "adoring",
    "affectionate",
    "agitated",
    "amazing",
    "angry",
    "awesome",
    "blissful",
    "bold",
    "boring",
    "brave",
    "charming",
    "clever",
    "cocky",
    "cool",
    "compassionate",
    "competent",
    "condescending",
    "confident",
    "cranky",
    "crazy",
    "dazzling",
    "determined",
    "distracted",
    "dreamy",
    "eager",
    "ecstatic",
    "elastic",
    "elated",
    "elegant",
    "eloquent",
    "epic",
    "fervent",
    "festive",
    "flamboyant",
    "focused",
    "friendly",
    "frosty",
    "gallant",
    "gifted",
    "goofy",
    "gracious",
    "happy",
    "hardcore",
    "heuristic",
    "hopeful",
    "hungry",
    "infallible",
    "inspiring",
    "jolly",
    "jovial",
    "keen",
    "kind",
    "laughing",
    "loving",
    "lucid",
    "magical",
    "mystifying",
    "modest",
    "musing",
    "naughty",
    "nervous",
    "nifty",
    "nostalgic",
    "objective",
    "optimistic",
    "peaceful",
    "pedantic",
    "pensive",
    "practical",
    "priceless",
    "quirky",
    "quizzical",
    "recursing",
    "relaxed",
    "reverent",
    "romantic",
    "sad",
    "serene",
    "sharp",
    "silly",
    "sleepy",
    "stoic",
    "stupefied",
    "suspicious",
    "sweet",
    "tender",
    "thirsty",
    "trusting",
    "unruffled",
    "upbeat",
    "vibrant",
    "vigilant",
    "vigorous",
    "wizardly",
    "wonderful",
    "xenodochial",
    "youthful",
    "zealous",
    "zen",
]


def random_name():
    """ Returns a random name "first name" & "last name """

    name = "%s_%s" % (choice(right_names), choice(left_names))

    return name



from io import BytesIO
from jumpscale.data.serializers.json import dumps
from .jsnacl import *


def payload_build(self, *args):
    """
    build a bytesIO buffer with all arguments serialized to somethign repeatable
    """
    buffer = BytesIO()
    for item in args:
        if not isinstance(item, bytes):
            if isinstance(item, str):
                item = item.encode()
            elif isinstance(item, int) or isinstance(item, float):
                item = str(item).encode()
            elif isinstance(item, dict):
                item = dumps(item).encode()
            else:
                raise ValueError(f"Got {item} supported types are bytes,str,int,float,dict")
        buffer.write(item)
    return buffer.getvalue()




class StorageInterface:
    def get(self, model, obj_id):
        pass

    def set(self, model, obj_id, value):
        pass

    def get_keys_in_model(self, model):
        pass

    def incr_id(self, model):
        pass

class IndexInterface:
    def __init__(self, bcdb_namespace):
        pass

    def get(self, model, index_prop, index_value):
        pass

    def set(self, model, index_prop, index_value, obj_id, old_value=None):
        pass

class IndexSetInterface:
    def __init__(self, bcdb_namespace):
        pass

    def get(self, model, index_prop, min, max):
        pass

    def set(self, model, obj):
        pass

class IndexTextInterface:
    def __init__(self, bcdb_namespace):
        pass

    def set(self, model, obj):
       pass

    def get(self, model, index_prop, pattern):
        pass 

class SerializerInterface:
    def loads(self, model, s):
        pass

    def dumps(self, model, data):
        pass


"""Config module is the single entry of configurations across the framework.

It allows
- resolving configurations paths for configuration directory`config_root`, or configuration file path `config_path`
- rebuildling default configurations or retrieving them (using `get_default_config`)
- Getting configurations using `get_default_config`
- Updating configuration using `update_config`



JS-NG> j.core.config
<module 'jumpscale.core.config' from '/home/ahmed/wspace/js/js-ng/jumpscale/core/config/__init__.py'>

## Getting where is the config.toml path
```
JS-NG> j.core.config.config_path
'/home/ahmed/.config/jumpscale/config.toml'
```

## Getting configuration directory
```
JS-NG> j.core.config.config_root
'/home/ahmed/.config/jumpscale'
```

## Getting default configurations of js-ng (used in case of no configuration found)
```
JS-NG> j.core.config.get_default_config()
{'debug': True, 'logging': {'redis': {'enabled': True, 'level': 15, 'max_size': 1000, 'dump': True, 'dump_dir': '/home/ahmed/.config/jumpscale/logs/redis'}, 'filesystem': {'enabled': True, 'level': 15, 'log_dir': '/home/ahmed/.config/jumpscale/logs/fs/log.txt', 'rotation': '5 MB'}}, 'alerts': {'enabled': True, 'level': 40}, 'ssh_key_path': '', 'private_key_path': '', 'stores': {'redis': {'hostname': 'localhost', 'port': 6379}, 'filesystem': {'path': '/home/ahmed/.config/jumpscale/secureconfig'}}, 'store': 'filesystem', 'threebot': {'default': ''}, 'explorer': {'default_url': 'https://explorer.testnet.grid.tf/explorer'}}
```

## Getting the current configurations
```
JS-NG> j.core.config.get_config()
{'debug': True, 'ssh_key_path': '', 'log_to_redis': False, 'log_to_files': True, 'log_level': 15, 'private_key_path': '/home/ahmed/.config/jumpscale/mykey.priv', 'secure_config_path': '/home/ahmed/.config/jumpscale/secureconfig', 'store': 'filesystem', 'logging': {'handlers': [{'sink': 'sys.stdout', 'format': '{time} - {message}', 'colorize': True, 'enqueue': True}, {'sink': '/home/ahmed/.config/jumpscale/logs/file_jumpscale.log', 'serialize': True, 'enqueue': True}], 'redis': {'enabled': True, 'level': 15, 'max_size': 1000, 'dump': True, 'dump_dir': '/home/ahmed/.config/jumpscale/logs/redis'}, 'filesystem': {'enabled': True, 'level': 15, 'log_dir': '/home/ahmed/.config/jumpscale/logs/fs/log.txt', 'rotation': '5 MB'}}, 'stores': {'redis': {'hostname': 'localhost', 'port': 6379}, 'filesystem': {'path': '/home/ahmed/.config/jumpscale/secureconfig'}}, 'alerts': {'enabled': True, 'level': 40}, 'threebot': {'default': ''}, 'explorer': {'default_url': 'https://explorer.testnet.grid.tf/explorer'}}
```

## Updating configurations

```
JS-NG> conf = j.core.config.get_default_config()
JS-NG> conf
{'debug': True, 'ssh_key_path': '', 'log_to_redis': False, 'log_to_files': True, 'log_level': 15, 'private_key_path': '/home/ahmed/.config/jumpscale/mykey.priv', 'secure_config_path': '/home/ahmed/.config/jumpscale/secureconfig', 'store': 'filesystem', 'logging': {'handlers': [{'sink': 'sys.stdout', 'format': '{time} - {message}', 'colorize': True, 'enqueue': True}, {'sink': '/home/ahmed/.config/jumpscale/logs/file_jumpscale.log', 'serialize': True, 'enqueue': True}], 'redis': {'enabled': True, 'level': 15, 'max_size': 1000, 'dump': True, 'dump_dir': '/home/ahmed/.config/jumpscale/logs/redis'}, 'filesystem': {'enabled': True, 'level': 15, 'log_dir': '/home/ahmed/.config/jumpscale/logs/fs/log.txt', 'rotation': '5 MB'}}, 'stores': {'redis': {'hostname': 'localhost', 'port': 6379}, 'filesystem': {'path': '/home/ahmed/.config/jumpscale/secureconfig'}}, 'alerts': {'enabled': True, 'level': 40}, 'threebot': {'default': ''}, 'explorer': {'default_url': 'https://explorer.testnet.grid.tf/explorer'}}

JS-NG> conf['favcolor'] = 'blue'
JS-NG> j.core.config.update_config(conf)
JS-NG> j.core.config.get_config()
{'debug': True, 'ssh_key_path': '', 'log_to_redis': False, 'log_to_files': True, 'log_level': 15, 'private_key_path': '/home/ahmed/.config/jumpscale/mykey.priv', 'secure_config_path': '/home/ahmed/.config/jumpscale/secureconfig', 'store': 'filesystem', 'favcolor': 'blue', 'logging': {'handlers': [{'sink': 'sys.stdout', 'format': '{time} - {message}', 'colorize': True, 'enqueue': True}, {'sink': '/home/ahmed/.config/jumpscale/logs/file_jumpscale.log', 'serialize': True, 'enqueue': True}], 'redis': {'enabled': True, 'level': 15, 'max_size': 1000, 'dump': True, 'dump_dir': '/home/ahmed/.config/jumpscale/logs/redis'}, 'filesystem': {'enabled': True, 'level': 15, 'log_dir': '/home/ahmed/.config/jumpscale/logs/fs/log.txt', 'rotation': '5 MB'}}, 'stores': {'redis': {'hostname': 'localhost', 'port': 6379}, 'filesystem': {'path': '/home/ahmed/.config/jumpscale/secureconfig'}}, 'alerts': {'enabled': True, 'level': 40}, 'threebot': {'default': ''}, 'explorer': {'default_url': 'https://explorer.testnet.grid.tf/explorer'}}
```

## Get/Set
you can use `j.core.config.get` and `j.core.config.set` to retrive values of keys or set the value of a key respectively.

e.g

```
JS-NG> j.core.config.get("favcolor")
'blue'

JS-NG> j.core.config.set("favcolor", "grey")
JS-NG> j.core.config.get("favcolor")
'grey'
```


"""

import os

import nacl.utils
import nacl.encoding
import pytoml as toml
from nacl.public import PrivateKey


__all__ = [
    "config_path",
    "config_root",
    "get_default_config",
    "get_config",
    "update_config",
    "Environment",
    "get",
    "set",
    "get_current_version",
]

__version__ = "0.1.0"


config_root = os.path.expanduser(os.path.join("~/.config", "jumpscale"))
config_path = os.path.join(config_root, "config.toml")


def get_current_version():
    return __version__


def get_default_config():
    """retrieves default configurations for plain jumpscale

    Returns:
        dict: default configuration
    """
    return {
        "debug": True,
        "shell": "ptpython",
        "logging": {
            "redis": {
                "enabled": True,
                "level": 15,
                "max_size": 1000,
                "dump": True,
                "dump_dir": os.path.join(config_root, "logs/redis"),
            },
            "filesystem": {
                "enabled": True,
                "level": 15,
                "log_dir": os.path.join(config_root, "logs/fs/log.txt"),
                "rotation": "5 MB",
            },
        },
        "alerts": {"enabled": True, "level": 40},
        "ssh_key_path": "",
        "private_key_path": "",
        "stores": {
            "redis": {"hostname": "localhost", "port": 6379},
            "filesystem": {"path": os.path.expanduser(os.path.join(config_root, "secureconfig"))},
            "whoosh": {"path": os.path.expanduser(os.path.join(config_root, "whoosh_indexes"))},
        },
        "factory": {"always_reload": False},
        "store": "filesystem",
        "threebot": {"default": "",},
        "explorer": {"default_url": "https://explorer.testnet.grid.tf/explorer",},
    }


def get_config():
    """Gets jumpscale configurations

    Returns:
        [dict] - toml loaded config of CONFIG_DIR/config.toml
    """
    with open(config_path, "r") as f:
        return toml.load(f)


def update_config(data):
    """Update jumpscale config with new data

    Arguments:
        data {dict} -- dict to update the config with.
    """
    with open(config_path, "w") as f:
        toml.dump(data, f)


def get(key):
    """ Retrives value from jumpscale config

    Arguments:
        key (str): the key you wish to retrieve
    """
    conf = get_config()
    return conf.get(key)


def set(key, val):
    """ Sets value in jumpscale config

    Arguments:
        key (str): the key you wish to update
        val: value to update with
    """
    conf = get_config()
    conf[key] = val
    update_config(conf)


def migrate_config():
    """add missing keys to current config from default"""

    default_config = get_default_config()
    current_config = get_config()

    def copy_missing(default, current):
        for key, value in default.items():
            if key not in current:
                current[key] = value
            elif isinstance(value, dict):
                copy_missing(value, current[key])

    copy_missing(default_config, current_config)
    update_config(current_config)


# Create default configurations file on loading.
if not os.path.exists(config_path):
    os.makedirs(os.path.dirname(config_path), exist_ok=True)
    with open(config_path, "w") as f:
        # don't ues mknod https://github.com/threefoldtech/js-ng/issues/182
        f.close()


def generate_key(basepath):
    hsk = PrivateKey.generate()
    hpk = hsk.public_key
    skpath = basepath + ".priv"
    pkpath = basepath + ".pub"
    with open(skpath, "wb") as f:
        f.write(hsk.encode(nacl.encoding.Base64Encoder()))
    with open(pkpath, "wb") as f:
        f.write(hpk.encode(nacl.encoding.Base64Encoder()))
    return skpath


class StoreTypeNotFound(Exception):
    pass


class Environment:
    def get_private_key_path(self):
        config = get_config()
        private_key_path = config["private_key_path"]
        return private_key_path

    def get_threebot_data(self):
        config = get_config()
        return config.get("threebot", {})

    def get_private_key(self):
        private_key_path = self.get_private_key_path()
        if not private_key_path:
            keypath = os.path.join(config_root, "privatekey")
            private_key_path = generate_key(keypath)
            config = get_config()
            config["private_key_path"] = private_key_path
            update_config(config)
        return open(private_key_path, "rb").read()

    def get_store_config(self, name):
        config = get_config()
        stores = config["stores"]
        if name not in stores:
            raise StoreTypeNotFound(f"'{name}' store is not found")
        return stores[name]

    def get_logging_config(self):
        return get_config()["logging"]


migrate_config()



import base64


def encode(s):
    """encode string with base64 algorithm

    Arguments:
        s (string) : the string will be encoded

    Returns:
        bytes : the encoded bytes
    """
    if isinstance(s, str):
        b = s.encode()
    else:
        b = s
    return base64.b64encode(b)


def decode(b):
    """decode base64 bytes to original obj

    Arguments:
        b (bytes) : the bytes will be decoded

    Returns:
         (string) : the decoded string
    """
    if isinstance(b, str):
        b = b.encode()
    return base64.b64decode(b)



import os
import re
import time
import sys
import traceback
import argparse
import requests

import inspect
import cgi

from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completion
from prompt_toolkit.shortcuts import print_formatted_text
from prompt_toolkit.eventloop.async_generator import AsyncGeneratorItem
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.styles import Style
from prompt_toolkit.formatted_text import HTML
from jumpscale import threesdk
from jumpscale.threesdk import settings
from jumpscale.core.exceptions.exceptions import JSException
from jumpscale.clients.docker.docker import DockerClient
from jumpscale.threesdk.threebot import ThreeBot, DEFAULT_IMAGE
from jumpscale.core.config import get_current_version


BASE_CONFIG_DIR = os.path.join(os.environ.get("HOME", "/root"), ".jsng")
HISTORY_FILENAME = os.path.join(BASE_CONFIG_DIR, "history.txt")

DEFAULT_TOOLBAR_MSG = "Welcome to 3sdk enter info for help"

style = Style.from_dict(
    {
        # User input (default text).
        "bottom-toolbar": "#ffffff bg:#333333",
        "default": "#aaaaaa",
        # Prompt.
    }
)


def get_binary_link():
    resp = requests.get("https://api.github.com/repos/threefoldtech/js-ng/releases/latest")
    resp = resp.json()
    # get versions
    download_link = ""
    version = resp["tag_name"]
    for platform in resp["assets"]:
        if sys.platform in platform["name"]:
            download_link = platform["browser_download_url"]
    return version, download_link


def update():
    print("checking for updates")
    latest_version, binary_link = get_binary_link()
    current_version = get_current_version()
    if latest_version != current_version:
        print(f"version: {latest_version} is available get it from {binary_link}")
        return
    docker_client = DockerClient()
    print("Checking for new docker image")
    docker_client.client.images.pull(f"{DEFAULT_IMAGE}:{latest_version}")
    print("Starting 3sdk containers")
    for container_name in os.listdir(os.path.expanduser("~/.config/jumpscale/containers")):
        ThreeBot.delete(container_name)
        ThreeBot.install(container_name)


def print_error(error):
    print_formatted_text(HTML("<ansired>{}</ansired>".format(cgi.html.escape(str(error)))))


def partition_line(line):
    def replacer(m):
        return m.group().replace(" ", "\0").strip("\"'")

    result = re.sub(r"""(['"]).*?\1""", replacer, line)
    parts = []
    for part in result.split():
        parts.append(part.replace("\0", " "))
    return parts


def noexpert_error(error):
    reports_location = f"{os.environ.get('HOME', os.environ.get('USERPROFILE', ''))}/sandbox/reports"
    error_file_location = f"{reports_location}/jsngreport_{time.strftime('%d%H%M%S')}.log"
    if not os.path.exists(reports_location):
        os.makedirs(reports_location)
    with open(error_file_location, "w") as f:
        f.write(str(error))
    err_msg = f"""Something went wrong. Please contact support at https://support.grid.tf/
Error report file has been created on your machine in this location: {error_file_location}
        """
    return err_msg


class Shell(Validator):
    def __init__(self):
        self._prompt = PromptSession()
        self.mode = None
        self.toolbarmsg = DEFAULT_TOOLBAR_MSG

    def get_completions_async(self, document, complete_event):
        text = document.current_line_before_cursor
        parts = partition_line(text)
        if not parts:
            root = None
            more = []
        else:
            root, more = parts[0], parts[1:]
        items = []
        if not root or not hasattr(threesdk, root):
            style = "bg:ansibrightblue"
            items += threesdk.__all__
            self.toolbarmsg = DEFAULT_TOOLBAR_MSG
        else:
            style = "bg:ansigreen"
            obj = getattr(threesdk, root)
            if not more or not hasattr(obj, more[0]):
                # complete object attributes
                self.toolbarmsg = threesdk._get_doc_line(obj.__doc__)
                for name, member in inspect.getmembers(obj, inspect.isroutine):
                    if not name.startswith("_"):
                        items.append(name)
                text = "" if not more else more[-1]
            else:
                # complete arguments
                func = getattr(obj, more[0])
                self.toolbarmsg = threesdk._get_doc_line(func.__doc__)
                style = "bg:ansired"
                for arg in inspect.getfullargspec(func).args:
                    field = arg + "="
                    if field in text:
                        continue
                    items.append(field)
                if len(more) > 1:
                    text = more[-1]
                else:
                    text = ""

        for item in items:
            if not item:
                continue
            if isinstance(item, Completion):
                item.start_position = -len(text)
            else:
                item = Completion(item, -len(text))
            regex = ".*".join(text)
            item.style = style
            if not text or re.search(regex, item.text):
                yield AsyncGeneratorItem(item)

    def bottom_toolbar(self):
        return [("class:bottom-toolbar", self.toolbarmsg)]

    def validate(self, document):
        text = document.current_line_before_cursor
        if not text:
            return
        root, *more = text.split(" ")
        submodule = getattr(threesdk, root, None)
        if not submodule:
            raise ValidationError(message=f"No such subcommand {root}")
        if not more and callable(submodule):
            func = root
        elif more:
            func = getattr(submodule, more[0], None)
            if not func:
                raise ValidationError(message=f"{root} has no command called {more[0]}")
        else:
            raise ValidationError(message="Invalid command")
        # TODO: validate args
        return

    def get_func_kwargs(self, cmd):
        parts = partition_line(cmd)
        root, extra = parts[0], parts[1:]
        module = getattr(threesdk, root)
        if inspect.isroutine(module):
            return module, self.get_kwargs(module, *extra)
        else:
            func = getattr(module, extra[0])
            return func, self.get_kwargs(func, *extra[1:])

    def get_kwargs(self, func, *args):
        funcspec = inspect.getfullargspec(func)
        kwargs = {}
        for arg in args:
            key, val = arg.split("=", 1)
            isbool = funcspec.annotations.get(key) is bool
            if isbool:
                if val:
                    val = val.lower() in ["y", "yes", "1", "true"]
                else:
                    val = True
            kwargs[key] = val
        return kwargs

    def execute(self, cmd):
        if not cmd.strip():
            return
        try:
            func, kwargs = self.get_func_kwargs(cmd)
            func(**kwargs)
        except JSException as e:
            if not settings.expert:
                print_error(str(e))
            else:
                print_error(traceback.format_exc())
        except Exception:
            if not settings.expert:
                print_error(noexpert_error(traceback.format_exc()))
            else:
                print_error(traceback.format_exc())

    def make_prompt(self):
        root = ("class:default", "3sdk>")
        while True:
            try:
                result = self.prompt([root])
                self.execute(result)
            except (EOFError, KeyboardInterrupt):
                sys.exit(0)

    def prompt(self, msg):
        return self._prompt.prompt(
            msg, completer=self, validator=self, style=style, bottom_toolbar=self.bottom_toolbar,
        )


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("--update", action="store_true", help="Update 3sdk")
    parser.add_argument("--expert", action="store_true", help="Run 3sdk in expert mode")
    args = parser.parse_args()
    settings.expert = args.expert

    if args.update:
        update()
    else:
        shell = Shell()
        shell.make_prompt()


if __name__ == "__main__":
    run()



from jumpscale.core.base import Base, fields
from jumpscale.loader import j
from gevent.pool import Pool
from bottle import Bottle, abort, request, response
from jumpscale.servers.gedis.server import GedisErrorTypes
from gevent.pywsgi import WSGIServer
from jumpscale.core.base import StoredFactory
import json


class GedisHTTPServer(Base):
    host = fields.String(default="127.0.0.1")
    port = fields.Integer(default=8000)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = Bottle()
        self._client = None
        self._app.route("/<package>/<actor>/<method>", ["GET", "POST"], self.handler)

    @property
    def client(self):
        if self._client is None:
            self._client = j.clients.gedis.get(self.instance_name)
            self._client.disable_deserialization = True
        return self._client

    def make_response(self, code, content):
        response.status = code
        response.content_type = 'application/json'
        return json.dumps(content)

    def handler(self, package, actor, method):
        actors = self.client.actors

        actor = getattr(actors, f"{package}_{actor}", None)
        if not actor:
            return self.make_response(400, {"error": "actor not found"})

        method = getattr(actor, method, None)
        if not method:
            return self.make_response(400, {"error": "method not found"})

        kwargs = request.json or dict()
        response = method(**kwargs)

        if not response.success:
            if response.error_type == GedisErrorTypes.NOT_FOUND:
                return self.make_response(404, {"error": response.error})

            elif response.error_type == GedisErrorTypes.BAD_REQUEST:
                return self.make_response(400, {"error": response.error})

            else:
                return self.make_response(500, {"error": response.error})

        return self.make_response(200, response.result)

    @property
    def gevent_server(self):
        return WSGIServer((self.host, self.port), self._app, spawn=Pool())

def export_module_as():
    return StoredFactory(GedisHTTPServer)



from .interfaces import *
from redis import Redis
import json
import sqlite3
from jumpscale.loader import j


class JSONSerializer(SerializerInterface):
    def loads(self, model, s):
        return model.load_obj_from_dict(json.loads(s))

    def dumps(self, model, data):
        return json.dumps(model.get_dict(data))


class RedisStorageClient(StorageInterface):
    def __init__(self, bcdb_namespace, host="localhost", port=6379, serializer=None):
        self.redis_client = Redis(host=host, port=port)
        self.serializer = serializer or JSONSerializer()
        self.bcdb_namespace = bcdb_namespace

    def get(self, model, obj_id):
        obj_str = self.redis_client.get(f"{self.bcdb_namespace}.{model.name}://{obj_id}")
        return self.serializer.loads(model, obj_str) if obj_str else None

    def set(self, model, obj_id, value):
        return self.redis_client.set(
            f"{self.bcdb_namespace}.{model.name}://{obj_id}", self.serializer.dumps(model, value)
        )

    def get_keys_in_model(self, model):
        pattern = f"{self.bcdb_namespace}.{model.name}://*"
        result = []
        cur, keys = self.redis_client.scan(cursor=0, match=pattern, count=2)
        result.extend(keys)
        while cur != 0:
            cur, keys = self.redis_client.scan(cursor=cur, match=pattern, count=2)
            result.extend(keys)
        return [self.get(model, int(x.split(b"://")[1])) for x in result]

    def incr_id(self, model):
        return self.redis_client.incr(f"{self.bcdb_namespace}.{model.name}.lastid")


class RedisIndexClient(IndexInterface):
    def __init__(self, bcdb_namespace, host="localhost", port=6379):
        self.redis_client = Redis(host=host, port=port)
        self.bcdb_namespace = bcdb_namespace

    def get(self, model, index_prop, index_value):
        res = self.redis_client.get(f"{self.bcdb_namespace}.indexer.{model.name}.{index_prop}://{index_value}")
        return int(res) if res else None

    def set(self, model, index_prop, index_value, obj_id, old_value=None):
        if old_value:
            self.redis_client.delete(f"{self.bcdb_namespace}.indexer.{model.name}.{index_prop}://{old_value}")
        return self.redis_client.set(f"{self.bcdb_namespace}.indexer.{model.name}.{index_prop}://{index_value}", obj_id)


class SQLiteIndexSetClient(IndexSetInterface):
    def __init__(self, bcdb_namespace):
        self.bcdb_namespace = bcdb_namespace

    def _create_if_not_exists(self, model):
        conn = sqlite3.connect(f"{self.bcdb_namespace}_index.db")
        c = conn.cursor()
        table_name = f"{model.name}"
        c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{table_name}'")
        if c.fetchone()[0] != 0:
            return None
        props = []
        props_str = ""
        for prop in model.schema.props.values():
            if prop.name != "id" and prop.index_key:
                prop_type = "INTEGER" if isinstance(prop.type, j.data.types.Integer) else "TEXT"
                props_str += f", {prop.name} {prop_type}"
                props.append(prop.name)

        c.execute(f"CREATE TABLE IF NOT EXISTS {table_name} (id int primary key{props_str})")
        for prop in props:
            c.execute(f"CREATE INDEX IF NOT EXISTS {table_name}_{prop}_index on {table_name}({prop})")
        conn.commit()
        conn.close()

    def get(self, model, index_prop, min, max):
        self._create_if_not_exists(model)
        conn = sqlite3.connect(f"{self.bcdb_namespace}_index.db")
        c = conn.cursor()
        table_name = f"{model.name}"
        c.execute(f"SELECT id FROM {table_name} WHERE {index_prop} >= ? and {index_prop} <= ?", (min, max))
        res = c.fetchall()
        conn.close()
        return res

    def set(self, model, obj):
        self._create_if_not_exists(model)
        conn = sqlite3.connect(f"{self.bcdb_namespace}_index.db")
        c = conn.cursor()
        table_name = f"{model.name}"
        props_str = "id"
        props = [obj.id]
        for prop in model.schema.props.values():
            if prop.name != "id" and prop.index_key:
                props_str += f", {prop.name}"
                props.append(getattr(obj, prop.name))

        c.execute(f"REPLACE INTO {table_name} ({props_str}) VALUES (?{', ?' * (len(props) - 1)})", props)
        conn.commit()
        conn.close()


class SonicIndexTextClient(IndexTextInterface):
    def __init__(self, bcdb_namespace):
        self.bcdb_namespace = bcdb_namespace
        try:
            self.sonic_client = j.clients.sonic.new(self.bcdb_namespace)
        except:
            self.sonic_client = j.clients.sonic.get(self.bcdb_namespace)

    def set(self, model, obj):
        for prop in model.schema.props.values():
            if prop.index_text:
                self.sonic_client.push(
                    self.bcdb_namespace, f"{model.name}_{prop.name}", str(obj.id), str(getattr(obj, prop.name))
                )

    def get(self, model, index_prop, pattern):
        return self.sonic_client.query(self.bcdb_namespace, f"{model.name}_{index_prop}", pattern)



"""This module helps with everything related to hashing strings, bytes with popular algorithms like, md5, sha256, sha384, sha512, blake2

```
JS-NG> j.data.hash.md5("abc")                                                                       
'900150983cd24fb0d6963f7d28e17f72'

JS-NG> j.data.hash.sha1("abc")                                                                      
'a9993e364706816aba3e25717850c26c9cd0d89d'

JS-NG> j.data.hash.sha224("abc")                                                                    
'23097d223405d8228642a477bda255b32aadbce4bda0b3f7e36c9da7'

JS-NG> j.data.hash.sha512("abc")                                                                    
'ddaf35a193617abacc417349ae20413112e6fa4e89a97ea20a9eeee64b55d39a2192992a274fc1a836ba3c23a3feebbd454
d4423643ce80e2a9ac94fa54ca49f'

JS-NG> j.data.hash.blake2s("abc")                                                                   
'508c5e8c327c14e2e1a72ba34eeb452f37458b209ed63a294d999b4c86675982'

JS-NG>
```  
"""

import hashlib
import os


def encode_string(obj, encode):
    if isinstance(obj, str):
        return obj.encode(encode)
    return obj


def get_list_files(dir_name):
    """returns a list of directories for all files in a root folder

    Arguments:
        dir_name (str) : the directory of the root folder

    Returns:
        all_files (list) : the list of directories for all files in the root folder
    """
    # create a list of file and sub directories
    # names in the given directory
    files_list = os.listdir(dir_name)
    all_files = list()
    # Iterate over all the entries
    for entry in files_list:
        # Create full path
        full_path = os.path.join(dir_name, entry)
        # If entry is a directory then get the list of files in this directory
        if os.path.isdir(full_path):
            all_files = all_files + get_list_files(full_path)
        else:
            all_files.append(full_path)

    return all_files


def md5(string, encode="utf-8"):
    """create a md5 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.md5(encode_string(string, encode)).hexdigest()


def sha1(string, encode="utf-8"):
    """create a sha1 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.sha1(encode_string(string, encode)).hexdigest()


def sha224(string, encode="utf-8"):
    """create a sha224 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.sha224(encode_string(string, encode)).hexdigest()


def sha384(string, encode="utf-8"):
    """create a sha384 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.sha384(encode_string(string, encode)).hexdigest()


def sha256(string, encode="utf-8"):
    """create a sha256 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.sha256(encode_string(string, encode)).hexdigest()


def sha512(string, encode="utf-8"):
    """create a sha512 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.sha512(encode_string(string, encode)).hexdigest()


def sha3_224(string, encode="utf-8"):
    """create a sha3_224 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.sha3_224(encode_string(string, encode)).hexdigest()


def sha3_256(string, encode="utf-8"):
    """create a sha3_256 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.sha3_256(encode_string(string, encode)).hexdigest()


def sha3_384(string, encode="utf-8"):
    """create a sha3_384 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.sha3_384(encode_string(string, encode)).hexdigest()


def sha3_512(string, encode="utf-8"):
    """create a sha3_512 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.sha3_512(encode_string(string, encode)).hexdigest()


def blake2s(string, encode="utf-8"):
    """create a blake2s hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.blake2s(encode_string(string, encode)).hexdigest()


def blake2b(string, encode="utf-8"):
    """create a blake2b hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.blake2b(encode_string(string, encode)).hexdigest()


def shake_128(string, encode="utf-8"):
    """create a shake_128 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.shake_128(encode_string(string, encode)).hexdigest(16)


def shake_256(string, encode="utf-8"):
    """create a shake_256 hash string for any string

    Arguments:
        string (str): the string need to be hashed

    Keyword Arguments:
        encode (str): the encoding for the string (default: {"utf-8"})

    Returns:
        list of byte: the hash bytes
    """
    return hashlib.shake_256(encode_string(string, encode)).hexdigest(16)


# TODO: review
def hash_alg(name, data, **kwargs):
    """create a hash object

    Arguments:
        name (str) : name of algorithm name
        data (str) : data need to be hashed

    Returns:
        list of byte : the hash bytes
    """
    return hashlib.new(name, data, **kwargs)


def hash_file(path, hash_type):
    """create hash string for a file

    Arguments:
        path (str) : the path for the file
        hash_type (str) : the type of the hash

    Returns:
        list of byte : the hash bytes
    """
    with open(path, "rb") as f:
        h = hashlib.new(hash_type)
        while True:
            data = f.read(2 ** 20)
            if not data:
                break
            h.update(data)
        return h


# TODO: review and allow ignore files e.g .swp, .. etc
def hash_directory(root_dir, hash_type):
    """create hash string list for the files in a folder

    Arguments:
        root_dir (str) : the dir for the root folder
        hash_type (str) : the type of the hash

    Returns:
        dict : the hashes dict, keys are full paths and values are hexdigests
    """
    hashes = {}
    for d in get_list_files(root_dir):
        hashes[d] = hash_file(d, hash_type).hexdigest()
    return hashes



from jumpscale.loader import j


DIR_PATH = j.sals.fs.dirname(j.sals.fs.realpath(__file__))
env = j.tools.jinja2.get_env(j.sals.fs.join_paths(DIR_PATH, "templates"))


def render_config_template(name, **kwargs):
    return env.get_template(f"{name}.conf").render(**kwargs)
